from __future__ import annotations
from arcgis.auth.tools import LazyLoader
from typing import Generator
from arcgis.geometry import Geometry
import copy
import datetime

try:
    import arcgis.graph._arcgisknowledge as _kgparser

    HAS_KG = True
except ImportError as e:
    HAS_KG = False
_isd = LazyLoader("arcgis._impl.common._isd")
from typing import List, Any


class KnowledgeGraph:
    """
    Provides access to the Knowledge Graph service data model and properties, as well as
    methods to search and query the graph.

    ==================     ====================================================================
    **Parameter**           **Description**
    ------------------     --------------------------------------------------------------------
    url                    Knowledge Graph service URL
    ------------------     --------------------------------------------------------------------
    gis                    an authenticated :class:`arcgis.gis.GIS` object.
    ==================     ====================================================================

    .. code-block:: python

        # Connect to a Knowledge Graph service:

        gis = GIS(url="url",username="username",password="password")

        knowledge_graph = KnowledgeGraph(url, gis=gis)

    """

    _gis = None
    _url = None
    _properties = None

    def __init__(self, url: str, *, gis=None):
        """initializer"""
        self._url = url
        self._gis = gis

    def _validate_import(self):
        if HAS_KG == False:
            raise ImportError(
                "An error occured with importing the Knowledge Graph libraries. Please ensure you "
                "are using Python 3.9, 3.10 or 3.11 on Windows or Linux platforms."
            )

    def _getInputQuantParams(self, inputQuantParams: dict):
        clientCoreQuantParams = _kgparser.InputQuantizationParameters()
        clientCoreQuantParams.xy_resolution = inputQuantParams["xyResolution"]
        clientCoreQuantParams.x_false_origin = inputQuantParams["xFalseOrigin"]
        clientCoreQuantParams.y_false_origin = inputQuantParams["yFalseOrigin"]
        clientCoreQuantParams.z_resolution = inputQuantParams["zResolution"]
        clientCoreQuantParams.z_false_origin = inputQuantParams["zFalseOrigin"]
        clientCoreQuantParams.m_resolution = inputQuantParams["mResolution"]
        clientCoreQuantParams.m_false_origin = inputQuantParams["mFalseOrigin"]
        return clientCoreQuantParams

    @classmethod
    def fromitem(cls, item):
        """Returns the Knowledge Graph service from an Item"""
        if item.type != "Knowledge Graph":
            raise ValueError(
                "Invalid item type, please provide a 'Knowledge Graph' item."
            )
        return cls(url=item.url, gis=item._gis)

    @property
    def properties(self) -> _isd.InsensitiveDict:
        """Returns the properties of the Knowledge Graph service"""
        if self._properties is None:
            resp = self._gis._con.get(self._url, {"f": "json"})
            self._properties = _isd.InsensitiveDict(resp)
        return self._properties

    def search(self, search: str, category: str = "both") -> List[dict]:
        """
        Allows for the searching of the properties of entities,
        relationships, or both in the graph using a full-text index.

        `Learn more about searching a knowledge graph <https://developers.arcgis.com/rest/services-reference/enterprise/kgs-graph-search.htm>`_

        ================    ===============================================================
        **Parameter**        **Description**
        ----------------    ---------------------------------------------------------------
        search              Required String. The search to perform on the Knowledge Graph.
        ----------------    ---------------------------------------------------------------
        category            Optional String.  The category is the location of the full
                            text search.  This can be isolated to either the `entities` or
                            the `relationships`.  The default is to look in `both`.

                            The allowed values are: both, entities, relationships,
                            both_entity_relationship, and meta_entity_provenance. Both and
                            both_entity_relationship are functionally the same.
        ================    ===============================================================

        .. note::
            Check the `service definition for the Knowledge Graph service <https://developers.arcgis.com/rest/services-reference/enterprise/kgs-hosted-server.htm>`_
            for valid values of category. Not all services support both and both_entity_relationship.

        .. code-block:: python

            #Perform a search on the knowledge graph
            search_result = knowledge_graph.search("cat")

            # Perform a search on only entities in the knowledge graph
            searchentities_result = knowledge_graph.search("cat", "entities")

        :return: List[list]

        """
        url = self._url + "/graph/search"
        cat_lu = {
            "both": _kgparser.esriNamedTypeCategory.both,
            "both_entity_relationship": _kgparser.esriNamedTypeCategory.both_entity_relationship,
            "relationships": _kgparser.esriNamedTypeCategory.relationship,
            "entities": _kgparser.esriNamedTypeCategory.entity,
            "meta_entity_provenance": _kgparser.esriNamedTypeCategory.meta_entity_provenance,
        }
        assert str(category).lower() in cat_lu.keys()
        r_enc = _kgparser.GraphSearchRequestEncoder()
        r_enc.search_query = search
        r_enc.return_geometry = True
        r_enc.max_num_results = self.properties["maxRecordCount"]
        r_enc.type_category_filter = cat_lu[category.lower()]
        r_enc.encode()
        error = r_enc.get_encoding_result().error
        if error.error_code != 0:
            raise Exception(error.error_message)
        query_dec = _kgparser.GraphQueryDecoder()

        session = self._gis._con._session
        response = session.post(
            url=url,
            params={"f": "pbf"},
            data=r_enc.get_encoding_result().byte_buffer,
            stream=True,
            headers={"Content-Type": "application/octet-stream"},
        )
        rows = []
        query_dec = _kgparser.GraphQueryDecoder()
        query_dec.data_model = self._datamodel
        for chunk in response.iter_content(8192):
            did_push = query_dec.push_buffer(chunk)
            while query_dec.next_row():
                rows.append(query_dec.get_current_row())
        return rows

    def update_search_index(self, adds: dict = None, deletes: dict = None) -> dict:
        """
        Allows users to add or delete search index properties for different entities and
        relationships from the graph's data model. Can only be existent properties for a given
        entity/relationship. Note that an empty dictionary result indicates success.

        =========================   ===============================================================
        **Parameter**                **Description**
        -------------------------   ---------------------------------------------------------------
        adds                        Optional dict. See below for structure. The properties to add
                                    to the search index, specified by entity/relationship.
        -------------------------   ---------------------------------------------------------------
        deletes                     Optional dict. See below for structure. The properties to
                                    delete from the search index, specified by entity/relationship.
        =========================   ===============================================================

        .. code-block:: python

            # example of an adds or deletes dictionary
            {
                "Entity1" : { "property_names": ["prop1", "prop2"]},
                "Entity2" : {"property_names": ["prop1"]},
                "RelationshipType1" : { "property_names": ["prop1", "prop2"]},
                "RelationshipType2" : {"property_names": ["prop1"]},
            }

        :return: A `dict`. Empty dict indicates success, errors will be returned in the dict.

        """

        self._validate_import()
        url = self._url + "/dataModel/searchIndex/update"
        params = {
            "f": "pbf",
            "token": self._gis._con.token,
        }
        headers = {"Content-Type": "application/octet-stream"}

        enc = _kgparser.GraphUpdateSearchIndexRequestEncoder()
        if adds:
            enc.insert_add_search_property(adds)
        if deletes:
            enc.insert_delete_search_property(deletes)

        enc.encode()
        enc_result = enc.get_encoding_result()
        error = enc_result.error
        if error.error_code != 0:
            raise Exception(error.error_message)

        session = self._gis._con._session
        response = session.post(
            url=url,
            params=params,
            data=enc_result.byte_buffer,
            stream=True,
            headers=headers,
        )

        content = response.content
        dec = _kgparser.GraphUpdateSearchIndexResponseDecoder()
        dec.decode(content)

        results = dec.get_results()
        return results

    def query(self, query: str) -> List[dict]:
        """
        Queries the Knowledge Graph using openCypher

        `Learn more about querying a knowledge graph <https://developers.arcgis.com/rest/services-reference/enterprise/kgs-graph-query.htm>`_

        ================    ===============================================================
        **Parameter**        **Description**
        ----------------    ---------------------------------------------------------------
        query               Required String. Allows you to return the entities and
                            relationships in a graph, as well as the properties of those
                            entities and relationships, by providing an openCypher query.
        ================    ===============================================================

        .. code-block:: python

            # Perform an openCypher query on the knowledge graph
            query_result = knowledge_graph.query("MATCH path = (n)-[r]-(n2) RETURN path LIMIT 5")


        :return: List[list]

        """
        self._validate_import()
        url = f"{self._url}/graph/query"
        params = {
            "f": "pbf",
            "token": self._gis._con.token,
            "openCypherQuery": query,
        }

        data = self._gis._con.get(url, params, return_raw_response=True, try_json=False)
        buffer_dm = data.content
        gqd = _kgparser.GraphQueryDecoder()
        gqd.push_buffer(buffer_dm)
        gqd.data_model = self._datamodel
        rows = []
        while gqd.next_row():
            r = gqd.get_current_row()
            rows.append(r)
        return rows

    def query_streaming(
        self,
        query: str,
        input_transform: dict[str, Any] = None,
        bind_param: dict[str, Any] = None,
        include_provenance: bool = False,
    ):
        """
        Query the graph using an openCypher query. Allows for more customization than the base
        `query()` function. Creates a generator of the query results, from which users can
        access each row or add them to a list. See below for example usage.


        ===================    ===============================================================
        **Parameter**           **Description**
        -------------------    ---------------------------------------------------------------
        query                  Required String. Allows you to return the entities and
                               relationships in a graph, as well as the properties of those
                               entities and relationships, by providing an openCypher query.
        -------------------    ---------------------------------------------------------------
        input_transform        Optional dict. Allows a user to specify custom quantization
                               parameters for input geometry, which dictate how geometries are
                               compressed and transferred to the server. Defaults to lossless
                               WGS84 quantization.
        -------------------    ---------------------------------------------------------------
        bind_param             Optional dict. The bind parameters used to filter
                               query results. Key of each pair is the string name for it,
                               which is how the parameter can be referenced in the query. The
                               value can be any "primitive" type value that may be found as
                               an attribute of an entity or relationship (e.g., string,
                               double, boolean, etc.), a list, an anonymous object (a dict),
                               or a geometry.

                               Anonymous objects and geometries can be passed
                               in as either their normal Python forms, or following the
                               format found in Knowledge Graph entries (containing an
                               "_objectType" key, and "_properties" for anonymous objects).

                               Note: Including bind parameters not used in the query will
                               cause queries to yield nothing on ArangoDB based services,
                               while Neo4j based services will still produce results.
        -------------------    ---------------------------------------------------------------
        include_provenance     Optional boolean. When `True`, provenance entities (metadata)
                               will be included in the query results. Defaults to `False`.
        ===================    ===============================================================

        .. code-block:: python

            # Get a list of all query results
            query_gen = knowledge_graph.query_streaming("MATCH path = (n)-[r]-(n2) RETURN path LIMIT 5")
            results = list(gen)

            # Grab one result at a time
            query_gen = knowledge_graph.query_streaming("MATCH path = (n)-[r]-(n2) RETURN path LIMIT 5")
            first_result = next(query_gen)
            second_result = next(query_gen)


        """

        self._validate_import()
        url = f"{self._url}/graph/query"
        params = {
            "f": "pbf",
            "token": self._gis._con.token,
        }
        headers = {"Content-Type": "application/octet-stream"}

        # initialize encoder
        r_enc = _kgparser.GraphQueryRequestEncoder()
        r_enc.open_cypher_query = query

        # set quant params
        if input_transform:
            quant_params = self._getInputQuantParams(input_transform)
        else:
            quant_params = _kgparser.InputQuantizationParameters.WGS84_lossless()
        r_enc.input_quantization_parameters = quant_params

        # set bind parameters
        if bind_param:

            def convert_to_properties(dictionary, last_key):
                if not isinstance(dictionary, dict):
                    return dictionary

                properties_dict = {}
                for key, value in dictionary.items():
                    if isinstance(value, dict):
                        if key != "_properties" and last_key == False:
                            properties_dict[key] = {
                                "_objectType": "object",
                                "_properties": convert_to_properties(value, False),
                            }
                        elif key != "properties" and last_key == True:
                            properties_dict[key] = convert_to_properties(value, False)
                        else:
                            properties_dict[key] = convert_to_properties(value, True)
                    else:
                        properties_dict[key] = value

                return properties_dict

            for k, v in bind_param.items():
                if isinstance(v, Geometry):
                    if "_objectType" not in v.keys():
                        copy_dict = copy.deepcopy(v)
                        copy_dict["_objectType"] = "geometry"
                        converted = _kgparser.from_value_object(copy_dict)
                    else:
                        converted = _kgparser.from_value_object(v)
                    r_enc.set_param_key_value(k, converted)

                elif isinstance(v, dict):
                    copy_dict = copy.deepcopy(v)
                    if "_properties" not in copy_dict.keys():
                        changed = {
                            "_objectType": "object",
                            "_properties": convert_to_properties(copy_dict, False),
                        }
                        converted = _kgparser.from_value_object(changed)
                    else:
                        if "_objectType" not in copy_dict.keys():
                            copy_dict["_objectType"] = "object"
                        copy_dict["_properties"] = convert_to_properties(
                            copy_dict["_properties"], True
                        )
                        converted = _kgparser.from_value_object(copy_dict)
                    r_enc.set_param_key_value(k, converted)

                elif isinstance(
                    v,
                    (
                        datetime.date,
                        datetime.time,
                        datetime.datetime,
                        datetime.timedelta,
                    ),
                ):
                    r_enc.set_param_key_value(k, v)
                else:
                    converted = _kgparser.from_value_object(v)
                    r_enc.set_param_key_value(k, converted)

        # set provenance behavior
        if include_provenance == True:
            r_enc.provenance_behavior = _kgparser.ProvenanceBehavior.include
        else:
            r_enc.provenance_behavior = _kgparser.ProvenanceBehavior.exclude

        r_enc.encode()
        error = r_enc.get_encoding_result().error
        if error.error_code != 0:
            raise Exception(error.error_message)
        query_dec = _kgparser.GraphQueryDecoder()

        session = self._gis._con._session
        response = session.post(
            url=url,
            params=params,
            data=r_enc.get_encoding_result().byte_buffer,
            stream=True,
            headers=headers,
        )

        for chunk in response.iter_content(8192):
            did_push = query_dec.push_buffer(chunk)
            while query_dec.next_row():
                yield query_dec.get_current_row()

    @property
    def _datamodel(self) -> object:
        """
        Returns the datamodel for the Knowledge Graph service
        """
        self._validate_import()
        url = f"{self._url}/dataModel/queryDataModel"
        params = {
            "f": "pbf",
        }
        r_dm = self._gis._con.get(
            url, params=params, return_raw_response=True, try_json=False
        )
        buffer_dm = r_dm.content
        dm = _kgparser.decode_data_model_from_protocol_buffer(buffer_dm)
        return dm

    @property
    def datamodel(self) -> dict:
        """
        Returns the datamodel for the Knowledge Graph service
        """
        self._validate_import()
        url = f"{self._url}/dataModel/queryDataModel"
        params = {
            "f": "pbf",
        }
        r_dm = self._gis._con.get(
            url, params=params, return_raw_response=True, try_json=False
        )
        buffer_dm = r_dm.content
        dm = _kgparser.decode_data_model_from_protocol_buffer(buffer_dm)
        return dm.to_value_object()

    def apply_edits(
        self,
        adds: list[dict[str, Any]] = [],
        updates: list[dict[str, Any]] = [],
        deletes: list[dict[str, Any]] = [],
        input_transform: dict[str, Any] = None,
        cascade_delete: bool = False,
        cascade_delete_provenance: bool = False,
    ) -> dict:
        """
        Allows users to add new graph entities/relationships, update existing
        entities/relationships, or delete existing entities/relationships. For details on how the
        dictionaries for each of these operations should be structured, please refer to the samples
        further below.

        .. note::
            objectid values are not supported in dictionaries for apply_edits

        =========================   ===============================================================
        **Parameter**                **Description**
        -------------------------   ---------------------------------------------------------------
        adds                        Optional list of dicts. The list of objects to add to the
                                    graph, represented in dictionary format.
        -------------------------   ---------------------------------------------------------------
        updates                     Optional list of dicts. The list of existent graph objects that
                                    are to be updated, represented in dictionary format.
        -------------------------   ---------------------------------------------------------------
        deletes                     Optional list of dicts. The list of existent objects to remove
                                    from the graph, represented in dictionary format.
        -------------------------   ---------------------------------------------------------------
        input_transform             Optional dict. Allows a user to specify custom quantization
                                    parameters for input geometry, which dictate how geometries are
                                    compressed and transferred to the server. Defaults to lossless
                                    WGS84 quantization.
        -------------------------   ---------------------------------------------------------------
        cascade_delete              Optional boolean. When `True`, relationships connected to
                                    entities that are being deleted will automatically be deleted
                                    as well. When `False`, these relationships must be deleted
                                    manually first. Defaults to `False`.
        -------------------------   ---------------------------------------------------------------
        cascade_delete_provenance   Optional boolean. When `True`, deleting entities/relationships
                                    or setting their property values to null will result in
                                    automatic deletion of associated provenance records. When
                                    `False`, `apply_edits()` will fail if there are provenance
                                    records connected to entities/relationships intended for
                                    deletion or having their properties set to null.
        =========================   ===============================================================

        .. code-block:: python

            # example of an add dictionary- include all properties
            {
                "_objectType": "entity",
                "_typeName": "Person",
                "_id": "{XXXXXXXX-XXXX-XXXX-XXXX-XXXXXXXXXXXXX}"
                "_properties": {
                    "name": "PythonAPILover",
                    "hometown": "Redlands",
                }
            }

            # update dictionary- include only properties being changed
            {
                "_objectType": "entity",
                "_typeName": "Person",
                "_id": "{XXXXXXXX-XXXX-XXXX-XXXX-XXXXXXXXXXXXX}"
                "_properties": {
                    "hometown": "Lisbon",
                }
            }

            # delete dictionary- pass a list of id's to be deleted
            {
                "_objectType": "entity",
                "_typeName": "Person",
                "_ids": ["{XXXXXXXX-XXXX-XXXX-XXXX-XXXXXXXXXXXXX}"]
            }

        :return: A `dict` showing the results of the edits.

        """

        url = self._url + "/graph/applyEdits"

        if input_transform:
            quant_params = self._getInputQuantParams(input_transform)
        else:
            quant_params = _kgparser.InputQuantizationParameters.WGS84_lossless()

        # now, make our encoder, and specify the edits to it
        enc = _kgparser.GraphApplyEditsEncoder(
            self._datamodel.spatial_reference,
            quant_params,
        )

        for edit in adds:
            enc.add(edit)
        for edit in updates:
            enc.update(edit)
        for edit in deletes:
            enc.delete_from_ids(edit)
        enc.cascade_delete = cascade_delete
        enc.cascade_delete_provenance = cascade_delete_provenance

        # encode and prepare for the post request
        enc.encode()
        res = enc.get_encoding_result()

        if res.error.error_code != 0:
            raise Exception(res.error.error_message)

        pbf_params = {
            "f": "pbf",
            "token": self._gis._con.token,
        }
        headers = {"Content-Type": "application/octet-stream"}

        # post and decode the response
        session = self._gis._con._session
        request_response = session.post(
            url,
            params=pbf_params,
            headers=headers,
            data=res.byte_buffer,
            stream=True,
        )
        apply_edits_response = request_response.content

        dec = _kgparser.GraphApplyEditsDecoder()
        dec.decode(apply_edits_response)
        results_dict = dec.get_results()

        return results_dict

    def named_object_type_adds(
        self,
        entity_types: list[dict[str, Any]] = [],
        relationship_types: list[dict[str, Any]] = [],
    ) -> dict:
        """
        Adds entity and relationship types to the data model

        `Learn more about adding named types to a knowledge graph <https://developers.arcgis.com/rest/services-reference/enterprise/kgs-datamodel-edit-namedtypes-add.htm>`_

        ==================  ===============================================================
        **Parameter**        **Description**
        ------------------  ---------------------------------------------------------------
        entity_types        Optional list of dicts. The list of entity types to add to the
                            data model, represented in dictionary format.
        ------------------  ---------------------------------------------------------------
        relationship_types  Optional list of dicts. The list of relationship types to add
                            to the data model, represented in dictionary format.
        ==================  ===============================================================

        .. code-block:: python

            # example of a named type to be added to the data model
            {
                "name": "Person",
                "alias": "Person",
                "role": "esriGraphNamedObjectRegular",
                "strict": False,
                "properties": {
                    "Name": {
                        "name": "Name",
                        "alias": "Name",
                        "fieldType": "esriFieldTypeString",
                        "editable": True,
                        "visible": True,
                        "required": False,
                        "isSystemMaintained": False,
                        "role": "esriGraphPropertyRegular"
                    },
                    "Nickname": {
                        "name": "Nickname",
                        "alias": "Nickname",
                        "fieldType": "esriFieldTypeString",
                        "editable": True,
                        "visible": True,
                        "required": False,
                        "isSystemMaintained": False,
                        "role": "esriGraphPropertyRegular"
                    }
                }
            }


        :return: A `dict` showing the results of the named type adds.

        """
        self._validate_import()
        url = f"{self._url}/dataModel/edit/namedTypes/add"

        r_enc = _kgparser.GraphNamedObjectTypeAddsRequestEncoder()
        for entity_type in entity_types:
            r_enc.add_entity_type(entity_type)
        for relationship_type in relationship_types:
            r_enc.add_relationship_type(relationship_type)

        r_enc.encode()
        error = r_enc.get_encoding_result().error
        if error.error_code != 0:
            raise Exception(error.error_message)
        r_dec = _kgparser.GraphNamedObjectTypeAddsResponseDecoder()

        session = self._gis._con._session
        response = session.post(
            url=url,
            params={"f": "pbf"},
            data=r_enc.get_encoding_result().byte_buffer,
            stream=True,
            headers={"Content-Type": "application/octet-stream"},
        )
        r_response = response.content

        r_dec.decode(r_response)
        results_dict = r_dec.get_results()

        return results_dict

    def named_object_type_update(
        self, type_name: str, named_type_update: dict[str, Any], mask: dict[str, Any]
    ) -> dict:
        """
        Updates an entity or relationship type in the data model

        `Learn more about updating named types in a knowledge graph <https://developers.arcgis.com/rest/services-reference/enterprise/kgs-datamodel-edit-namedtypes-type-update.htm>`_

        =================   ===============================================================
        **Parameter**        **Description**
        -----------------   ---------------------------------------------------------------
        type_name           Required string. The named type to be updated.
        -----------------   ---------------------------------------------------------------
        named_type_update   Required dict. The entity or relationship type to be updated,
                            represented in dictionary format.
        -----------------   ---------------------------------------------------------------
        mask                Required dict. A dictionary representing the properties of the
                            named type to be updated.
        =================   ===============================================================

        .. code-block:: python

            # example of a named type to be updated
            {
                "name": "Person",
                "alias": "Person",
                "role": "esriGraphNamedObjectRegular",
                "strict": False
            }

            # update the named type's alias:
            {
                "update_alias": True
            }
            # OR
            {
                "update_name": False,
                "update_alias": True,
                "update_role": False,
                "update_strict": False
            }


        :return: A `dict` showing the results of the named type update.

        """
        self._validate_import()
        url = f"{self._url}/dataModel/edit/namedTypes/{type_name}/update"

        data_model = self._datamodel
        entity_type = data_model.query_entity_type(type_name)
        relationship_type = data_model.query_relationship_type(type_name)

        r_enc = _kgparser.GraphNamedObjectTypeUpdateRequestEncoder()
        if entity_type is not None:
            r_enc.update_entity_type(named_type_update, mask)
        elif relationship_type is not None:
            r_enc.update_relationship_type(named_type_update, mask)

        r_enc.encode()
        error = r_enc.get_encoding_result().error
        if error.error_code != 0:
            raise Exception(error.error_message)
        r_dec = _kgparser.GraphNamedObjectTypeUpdateResponseDecoder()

        session = self._gis._con._session
        response = session.post(
            url=url,
            params={"f": "pbf"},
            data=r_enc.get_encoding_result().byte_buffer,
            stream=True,
            headers={"Content-Type": "application/octet-stream"},
        )
        r_response = response.content

        r_dec.decode(r_response)
        results_dict = r_dec.get_results()

        return results_dict

    def named_object_type_delete(self, type_name: str) -> dict:
        """
        Deletes an entity or relationship type in the data model

        `Learn more about deleting named types in a knowledge graph <https://developers.arcgis.com/rest/services-reference/enterprise/kgs-datamodel-edit-namedtypes-type-delete.htm>`_

        ================    ===============================================================
        **Parameter**        **Description**
        ----------------    ---------------------------------------------------------------
        type_name           Required string. The named type to be deleted.
        ================    ===============================================================

        .. code-block:: python

            # Delete a named type in the data model
            delete_result = knowledge_graph.named_object_type_delete("Person")


        :return: A `dict` showing the results of the named type delete.

        """
        self._validate_import()
        url = f"{self._url}/dataModel/edit/namedTypes/{type_name}/delete"

        r_dec = _kgparser.GraphNamedObjectTypeUpdateResponseDecoder()

        session = self._gis._con._session
        response = session.post(
            url=url,
            params={"f": "pbf"},
            stream=True,
            headers={"Content-Type": "application/octet-stream"},
        )
        r_response = response.content

        r_dec.decode(r_response)
        results_dict = r_dec.get_results()

        return results_dict

    def graph_property_adds(
        self, type_name: str, graph_properties: list[dict[str, Any]]
    ) -> dict:
        """
        Adds properties to a named type in the data model

        `Learn more about adding properties in a knowledge graph <https://developers.arcgis.com/rest/services-reference/enterprise/kgs-datamodel-edit-namedtypes-type-fields-add.htm>`_

        ================    ===============================================================
        **Parameter**        **Description**
        ----------------    ---------------------------------------------------------------
        type_name           Required string. The entity or relationship type to which the
                            properties will be added.
        ----------------    ---------------------------------------------------------------
        graph_properties    Required list of dicts. The list of properties to add
                            to the named type, represented in dictionary format.
        ================    ===============================================================

        .. code-block:: python

            # example of a shape property to be added to a named type
            {
                "name": "MyPointGeometry",
                "alias": "MyPointGeometry",
                "fieldType": "esriFieldTypeGeometry",
                "geometryType": "esriGeometryPoint",
                "hasZ": False,
                "hasM": False,
                "nullable": True,
                "editable": True,
                "visible": True,
                "required": False,
                "isSystemMaintained": False,
                "role": "esriGraphPropertyRegular"
            }

            # example of an integer property to be added to a named type
            {
                "name": "MyInt",
                "alias": "MyInt",
                "fieldType": "esriFieldTypeInteger",
                "nullable": True,
                "editable": True,
                "defaultValue": 123,
                "visible": True,
                "required": False,
                "isSystemMaintained": False,
                "role": "esriGraphPropertyRegular",
                "domain": "MyIntegerDomain"
            }


        :return: A `dict` showing the results of the property adds.

        """
        self._validate_import()
        url = f"{self._url}/dataModel/edit/namedTypes/{type_name}/fields/add"

        r_enc = _kgparser.GraphPropertyAddsRequestEncoder()
        for prop in graph_properties:
            r_enc.add_property(prop)

        r_enc.encode()
        error = r_enc.get_encoding_result().error
        if error.error_code != 0:
            raise Exception(error.error_message)
        r_dec = _kgparser.GraphPropertyAddsResponseDecoder()

        session = self._gis._con._session
        response = session.post(
            url=url,
            params={"f": "pbf"},
            data=r_enc.get_encoding_result().byte_buffer,
            stream=True,
            headers={"Content-Type": "application/octet-stream"},
        )
        r_response = response.content

        r_dec.decode(r_response)
        results_dict = r_dec.get_results()

        return results_dict

    def graph_property_update(
        self,
        type_name: str,
        property_name: str,
        graph_property: dict[str, Any],
        mask: dict[str, Any],
    ) -> dict:
        """
        Updates a property for a named type in the data model

        `Learn more about updating properties in a knowledge graph <https://developers.arcgis.com/rest/services-reference/enterprise/kgs-datamodel-edit-namedtypes-type-fields-update.htm>`_

        ================    ===============================================================
        **Parameter**        **Description**
        ----------------    ---------------------------------------------------------------
        type_name           Required string. The entity or relationship type containing
                            the property to be updated.
        ----------------    ---------------------------------------------------------------
        property_name       Required string. The property to be updated.
        ----------------    ---------------------------------------------------------------
        graph_property      Required dict. The graph property to be updated,
                            represented in dictionary format.
        ----------------    ---------------------------------------------------------------
        mask                Required dict. A dictionary representing the properties of the
                            field to be updated.
        ================    ===============================================================

        .. code-block:: python

            # example of a shape property to be updated
            {
                "name": "MyPointGeometry",
                "alias": "MyPointGeometry",
                "fieldType": "esriFieldTypeGeometry",
                "geometryType": "esriGeometryPoint",
                "hasZ": False,
                "hasM": False,
                "nullable": True,
                "editable": True,
                "visible": True,
                "required": False,
                "isSystemMaintained": False,
                "role": "esriGraphPropertyRegular"
            }

            # example: update the property's alias
            {
                "update_alias": True
            }
            # OR
            {
                "update_name": False,
                "update_alias": True,
                "update_field_type": False,
                "update_geometry_type": False,
                "update_default_value": False,
                "update_nullable": False,
                "update_editable": False,
                "update_visible": False,
                "update_required": False,
                "update_has_z": False,
                "update_has_m": False,
                "update_domain:" False
            }


        :return: A `dict` showing the results of the property update.

        """
        self._validate_import()
        url = f"{self._url}/dataModel/edit/namedTypes/{type_name}/fields/update"

        r_enc = _kgparser.GraphPropertyUpdateRequestEncoder()
        r_enc.update_property(graph_property, mask)
        r_enc.name = property_name

        r_enc.encode()
        error = r_enc.get_encoding_result().error
        if error.error_code != 0:
            raise Exception(error.error_message)
        r_dec = _kgparser.GraphPropertyUpdateResponseDecoder()

        session = self._gis._con._session
        response = session.post(
            url=url,
            params={"f": "pbf"},
            data=r_enc.get_encoding_result().byte_buffer,
            stream=True,
            headers={"Content-Type": "application/octet-stream"},
        )
        r_response = response.content

        r_dec.decode(r_response)
        results_dict = r_dec.get_results()

        return results_dict

    def graph_property_delete(self, type_name: str, property_name: str) -> dict:
        """
        Delete a property for a named type in the data model

        `Learn more about deleting properties in a knowledge graph <https://developers.arcgis.com/rest/services-reference/enterprise/kgs-datamodel-edit-namedtypes-type-fields-delete.htm>`_

        ================    ===============================================================
        **Parameter**        **Description**
        ----------------    ---------------------------------------------------------------
        type_name           Required string. The entity or relationship type containing
                            the property to be deleted.
        ----------------    ---------------------------------------------------------------
        property_name       Required string. The property to be deleted.
        ================    ===============================================================

        .. code-block:: python

            # Delete a named type's property in the data model
            delete_result = knowledge_graph.graph_property_delete("Person", "Address")


        :return: A `dict` showing the results of the property delete.

        """
        self._validate_import()
        url = f"{self._url}/dataModel/edit/namedTypes/{type_name}/fields/delete"

        r_enc = _kgparser.GraphPropertyDeleteRequestEncoder()
        r_enc.name = property_name

        r_enc.encode()
        error = r_enc.get_encoding_result().error
        if error.error_code != 0:
            raise Exception(error.error_message)
        r_dec = _kgparser.GraphPropertyDeleteResponseDecoder()

        session = self._gis._con._session
        response = session.post(
            url=url,
            params={"f": "pbf"},
            data=r_enc.get_encoding_result().byte_buffer,
            stream=True,
            headers={"Content-Type": "application/octet-stream"},
        )
        r_response = response.content

        r_dec.decode(r_response)
        results_dict = r_dec.get_results()

        return results_dict



try:
    import os, sys, json
    import numpy as np
    import pandas as pd
    import torch
    from torch import tensor
    import torch.nn as nn
    import math
    from . import util

    HAS_TORCH = True
except Exception as e:
    HAS_TORCH = False
import arcgis
from arcgis.learn import AutoML

try:
    import arcpy
except:
    pass


class ChildImageClassifier:
    def initialize(self, model, model_as_file):
        if not HAS_TORCH:
            raise Exception(
                "Could not find the required deep learning dependencies. Ensure you have installed the required dependent libraries. See https://developers.arcgis.com/python/guide/deep-learning/"
            )

        if arcpy.env.processorType == "GPU" and torch.cuda.is_available():
            self.device = torch.device("cuda")
            arcgis.env._processorType = "GPU"
        else:
            self.device = torch.device("cpu")
            arcgis.env._processorType = "CPU"

        if model_as_file:
            with open(model, "r") as f:
                self.json_info = json.load(f)
        else:
            self.json_info = json.load(model)

        model_path = self.json_info["ResultsPath"]
        if model_as_file and not os.path.isabs(model_path):
            model_path = os.path.abspath(
                os.path.join(os.path.dirname(model), model_path)
            )

        self.automl = AutoML.from_model(emd_path=model)
        self._learnmodel = self.automl
        # self.model = self.pix2pix_hd.learn.model.to(self.device)
        # self.model.eval()

    def getParameterInfo(self, required_parameters):
        band_cnt = 0
        for col in self.json_info["_raster_field_variables"]:
            required_parameters.extend(
                [
                    {
                        "name": f"{col}",
                        "dataType": "numeric",
                        "value": band_cnt,
                        "required": True,
                        "displayName": "Band Value",
                        "description": "mapping of band",
                    },
                ]
            )
            band_cnt = band_cnt + 1

        return required_parameters

    def getConfiguration(self, **scalars):
        self.scalars = scalars

        return {"fixedTileSize": 1}

    def updatePixels(self, tlc, shape, props, **pixelBlocks):
        image = pixelBlocks["raster_pixels"].astype(np.float32)  # change to int

        cols = {}

        for col in self.json_info["_raster_field_variables"]:
            cols[f"{col}"] = image[int(self.scalars.get("{}".format(col)))].flatten()

        raster_data = cols

        processed_data = []

        length_values = len(raster_data[list(raster_data.keys())[0]])
        for i in range(length_values):
            processed_row = []
            for raster_name in sorted(raster_data.keys()):
                processed_row.append(raster_data[raster_name][i])
            processed_data.append(processed_row)

        processed_df = pd.DataFrame(
            data=np.array(processed_data), columns=sorted(raster_data)
        )
        processed_numpy = self.automl._data._process_data(processed_df, fit=False)
        predictions = self.automl._predict(processed_numpy)

        predictions = np.array(
            predictions.reshape([image.shape[1], image.shape[2]]),
            dtype="float64",
        )

        predictions = np.expand_dims(predictions, axis=0)

        return predictions



from logging import exception
from fastai.vision import (
    ItemBase,
    ItemList,
    Tensor,
    ImageList,
    Tuple,
    Path,
    get_transforms,
    random,
    open_image,
    Image,
    math,
    plt,
    torch,
    Learner,
    partial,
    optim,
    ifnone,
    image2np,
)
from .._utils.common import (
    ArcGISImageList,
    ArcGISMSImage,
    get_top_padding,
    get_nbatches,
)
from pathlib import Path
import torch.nn as nn
import torch.nn.functional as F
from torchvision import models
import torch
from scipy import linalg
from torch.nn.functional import adaptive_avg_pool2d
import numpy as np
from .._utils.cyclegan import get_activations, InceptionV3
import os
import json


class ImageTuple(ItemBase):
    def __init__(self, img1, img2, img3):
        self.img1, self.img2, self.img3 = img1, img2, img3
        self.obj, self.data = (
            (img1, img2, img3),
            [-1 + 2 * img1.data, -1 + 2 * img2.data, -1 + 2 * img3.data],
        )
        self.data2 = [-1 + 2 * img3.data, -1 + 2 * img2.data, -1 + 2 * img1.data]
        self.shape = img1.shape

    def apply_tfms(self, tfms, **kwargs):
        self.img1 = self.img1.apply_tfms(tfms, **kwargs)
        self.img2 = self.img2.apply_tfms(tfms, **kwargs)
        self.img3 = self.img3.apply_tfms(tfms, **kwargs)
        return self

    def to_one(self):
        return Image(
            0.5
            + torch.cat(
                [
                    self.data[0][:3, :, :],
                    self.data[1][:3, :, :],
                    self.data[2][:3, :, :],
                ],
                2,
            )
            / 2
        )

    def to_one_pred(self):
        return Image(0.5 + (self.data2[0]) / 2)

    def __repr__(self):
        return f"{self.__class__.__name__}{(self.img1.shape, self.img2.shape, self.img3.shape)}"


class TargetTupleList(ItemList):
    def reconstruct(self, t: Tensor):
        if len(t.size()) == 0:
            return t
        return ImageTuple(
            Image(t[0] / 2 + 0.5), Image(t[1] / 2 + 0.5), Image(t[2] / 2 + 0.5)
        )


_batch_stats_a = None
_batch_stats_b = None
_batch_stats_c = None


class ImageTupleListMS2(ArcGISImageList):
    _label_cls = TargetTupleList

    def __init__(
        self,
        items,
        itemsB=None,
        itemsC=None,
        itemsB_valid=None,
        itemsC_valid=None,
        **kwargs,
    ):
        self.itemsB = itemsB
        self.itemsC = itemsC
        self.itemsB_valid = itemsB_valid
        self.itemsC_valid = itemsC_valid
        super().__init__(items, **kwargs)

    def new(self, items, **kwargs):
        return super().new(
            items,
            itemsB=self.itemsB,
            itemsC=self.itemsC,
            itemsB_valid=self.itemsB_valid,
            itemsC_valid=self.itemsC_valid,
            **kwargs,
        )

    def get(self, i):
        if len(self.items) == len(self.itemsB):
            img1 = super().get(i)
            fn1 = self.itemsB[i]
            img2 = ArcGISMSImage.open(fn1)
            fn = self.itemsC[i]
            img3 = ArcGISMSImage.open(fn)

            max_of = np.max([img1.shape[0], img2.shape[0], img3.shape[0]])

            if max_of != img1.shape[0]:
                cont = []
                last_tile = np.expand_dims(img1.data[img1.shape[0] - 1, :, :], 0)
                res = abs(max_of - img1.shape[0])
                for i in range(res):
                    img1 = ArcGISMSImage(
                        torch.tensor(np.concatenate((img1.data, last_tile), axis=0))
                    )
            if max_of != img2.shape[0]:
                cont = []
                last_tile = np.expand_dims(img2.data[img2.shape[0] - 1, :, :], 0)
                res = abs(max_of - img2.shape[0])
                for i in range(res):
                    img2 = ArcGISMSImage(
                        torch.tensor(np.concatenate((img2.data, last_tile), axis=0))
                    )
            if max_of != img3.shape[0]:
                cont = []
                last_tile = np.expand_dims(img3.data[img3.shape[0] - 1, :, :], 0)
                res = abs(max_of - img3.shape[0])
                for i in range(res):
                    img3 = ArcGISMSImage(
                        torch.tensor(np.concatenate((img3.data, last_tile), axis=0))
                    )
            if max_of == 1:
                last_tile1, last_tile2, last_tile3 = (
                    np.expand_dims(img1.data[0, :, :], 0),
                    np.expand_dims(img2.data[0, :, :], 0),
                    np.expand_dims(img3.data[0, :, :], 0),
                )
                for i in range(2):
                    img1 = ArcGISMSImage(
                        torch.tensor(np.concatenate((img1.data, last_tile1), axis=0))
                    )
                    img2 = ArcGISMSImage(
                        torch.tensor(np.concatenate((img2.data, last_tile2), axis=0))
                    )
                    img3 = ArcGISMSImage(
                        torch.tensor(np.concatenate((img3.data, last_tile3), axis=0))
                    )
        else:
            img1 = super().get(i)
            fn1 = self.itemsB_valid[i]
            img2 = ArcGISMSImage.open(fn1)
            fn = self.itemsC_valid[i]
            img3 = ArcGISMSImage.open(fn)
            max_of = np.max([img1.shape[0], img2.shape[0], img3.shape[0]])

            if max_of != img1.shape[0]:
                cont = []
                last_tile = np.expand_dims(img1.data[img1.shape[0] - 1, :, :], 0)
                res = abs(max_of - img1.shape[0])
                for i in range(res):
                    img1 = ArcGISMSImage(
                        torch.tensor(np.concatenate((img1.data, last_tile), axis=0))
                    )
            if max_of != img2.shape[0]:
                cont = []
                last_tile = np.expand_dims(img2.data[img2.shape[0] - 1, :, :], 0)
                res = abs(max_of - img2.shape[0])
                for i in range(res):
                    img2 = ArcGISMSImage(
                        torch.tensor(np.concatenate((img2.data, last_tile), axis=0))
                    )
            if max_of != img3.shape[0]:
                cont = []
                last_tile = np.expand_dims(img3.data[img3.shape[0] - 1, :, :], 0)
                res = abs(max_of - img3.shape[0])
                for i in range(res):
                    img3 = ArcGISMSImage(
                        torch.tensor(np.concatenate((img3.data, last_tile), axis=0))
                    )
            if max_of == 1:
                last_tile1, last_tile2, last_tile3 = (
                    np.expand_dims(img1.data[0, :, :], 0),
                    np.expand_dims(img2.data[0, :, :], 0),
                    np.expand_dims(img3.data[0, :, :], 0),
                )
                for i in range(2):
                    img1 = ArcGISMSImage(
                        torch.tensor(np.concatenate((img1.data, last_tile1), axis=0))
                    )
                    img2 = ArcGISMSImage(
                        torch.tensor(np.concatenate((img2.data, last_tile2), axis=0))
                    )
                    img3 = ArcGISMSImage(
                        torch.tensor(np.concatenate((img3.data, last_tile3), axis=0))
                    )

        global _batch_stats_a
        global _batch_stats_b
        global _batch_stats_c

        img1_scaled = _tensor_scaler_tfm(
            img1.data,
            min_values=_batch_stats_a["band_min_values"],
            max_values=_batch_stats_a["band_max_values"],
            mode="minmax",
        )
        img2_scaled = _tensor_scaler_tfm(
            img2.data,
            min_values=_batch_stats_b["band_min_values"],
            max_values=_batch_stats_b["band_max_values"],
            mode="minmax",
        )
        img3_scaled = _tensor_scaler_tfm(
            img3.data,
            min_values=_batch_stats_c["band_min_values"],
            max_values=_batch_stats_c["band_max_values"],
            mode="minmax",
        )

        self.img1_scaled = ArcGISMSImage(img1_scaled)
        self.img2_scaled = ArcGISMSImage(img2_scaled)
        self.img3_scaled = ArcGISMSImage(img3_scaled)

        return ImageTuple(self.img1_scaled, self.img2_scaled, self.img3_scaled)

    def show(self, i, axes=None, rgb_bands=None):
        if rgb_bands is None:
            rgb_bands = [0, 1, 2]
        self[i]
        if axes is None:
            _, axes = plt.subplots(1, 3, figsize=(15, 7))

        self.img1_scaled.show(axes[0], rgb_bands=rgb_bands)
        self.img2_scaled.show(axes[1], rgb_bands=rgb_bands)
        self.img3_scaled.show(axes[2], rgb_bands=rgb_bands)

    def reconstruct(self, t: Tensor):
        if len(t) == 3:
            return ImageTuple(
                ArcGISMSImage(t[0] / 2 + 0.5),
                ArcGISMSImage(t[1] / 2 + 0.5),
                ArcGISMSImage(t[2] / 2 + 0.5),
            )
        else:
            return ImageTuple(
                ArcGISMSImage(t[0] / 2 + 0.5),
                ArcGISMSImage(t[1] / 2 + 0.5),
                ArcGISMSImage(t[1] / 2 + 0.5),
            )

    @classmethod
    def from_folders(
        cls,
        path,
        folderA,
        folderB,
        folderC,
        batch_stats_a,
        batch_stats_b,
        batch_stats_c,
        **kwargs,
    ):
        itemsB = ImageList.from_folder(folderB).items
        itemsC = ImageList.from_folder(folderC).items
        res = super().from_folder(
            folderA,
            itemsB=itemsB,
            itemsC=itemsC,
            itemsB_valid=itemsB,
            itemsC_valid=itemsC,
            **kwargs,
        )
        res.path = path

        global _batch_stats_a
        global _batch_stats_b
        global _batch_stats_c

        _batch_stats_a = batch_stats_a
        _batch_stats_b = batch_stats_b
        _batch_stats_c = batch_stats_c
        return res

    def split_by_idxs(self, train_idx, valid_idx):
        "Split the data between `train_idx` and `valid_idx`."

        train_idx_alt = []
        for i in range(train_idx.shape[0]):
            if train_idx[i] < len(self.itemsB):
                train_idx_alt.append(train_idx[i])

        valid_idx_alt = []
        for j in range(valid_idx.shape[0]):
            if valid_idx[j] < len(self.itemsB):
                valid_idx_alt.append(valid_idx[j])

        self.itemsB_valid = self.itemsB_valid[valid_idx_alt]
        self.itemsC_valid = self.itemsC_valid[valid_idx_alt]
        self.itemsB = self.itemsB[train_idx_alt]
        self.itemsC = self.itemsC[train_idx_alt]
        return self.split_by_list(self[train_idx_alt], self[valid_idx_alt])

    def show_xys(self, xs, ys, figsize: Tuple[int, int] = (12, 6), **kwargs):
        "Show the `xs` and `ys` on a figure of `figsize`. `kwargs` are passed to the show method."

        rows = int(math.sqrt(len(xs)))
        fig, axs = plt.subplots(rows, rows, figsize=figsize)
        for i, ax in enumerate(axs.flatten() if rows > 1 else [axs]):
            xs[i].to_one().show(ax=ax, **kwargs)
        plt.tight_layout()

    def show_xyzs(self, xs, ys, zs, figsize: Tuple[int, int] = None, **kwargs):
        """Show `xs` (inputs), `ys` (targets) and `zs` (predictions) on a figure of `figsize`.
        `kwargs` are passed to the show method."""

        figsize = ifnone(figsize, (12, 3 * len(xs)))
        fig, axs = plt.subplots(len(xs), 2, figsize=figsize)
        if axs.ndim == 1:  # fix for rows=1
            axs = axs.reshape(1, 2)
        fig.suptitle("Images / Ground truth / Predictions", weight="bold", size=14)
        for i, (x, z) in enumerate(zip(xs, zs)):
            x.to_one().show(ax=axs[i, 0], **kwargs)
            z.to_one_pred().show(ax=axs[i, 1], **kwargs)


def calculate_activation_statistics(batch_size, data_len, batch_list):
    act = get_activations(batch_size, data_len, batch_list)
    mu = np.mean(act, axis=0)
    sigma = np.cov(act, rowvar=False)
    return mu, sigma


def _tensor_scaler_tfm(tensor_batch, min_values, max_values, mode="minmax"):
    from .._data import _tensor_scaler

    x = tensor_batch
    if x.shape[0] > min_values.shape[0]:
        res = x.shape[0] - min_values.shape[0]
        last_val = torch.tensor([min_values[min_values.shape[0] - 1]])
        for i in range(res):
            min_values = torch.tensor(np.concatenate((min_values, last_val), axis=0))

    if x.shape[0] > max_values.shape[0]:
        res = x.shape[0] - max_values.shape[0]
        last_val = torch.tensor([max_values[max_values.shape[0] - 1]])
        for i in range(res):
            max_values = torch.tensor(np.concatenate((max_values, last_val), axis=0))
    max_values = max_values.view(-1, 1, 1).to(x.device)
    min_values = min_values.view(-1, 1, 1).to(x.device)
    x = _tensor_scaler(x, min_values, max_values, mode, create_view=False)
    return x


def _batch_stats_json(
    path, img_list, norm_pct, stats_file_name="esri_normalization_stats.json"
):
    from .._data import _get_batch_stats

    if len(img_list) < 300:
        norm_pct = 1

    dummy_stats = {
        "batch_stats_for_norm_pct_0": {
            "band_min_values": None,
            "band_max_values": None,
            "band_mean_values": None,
            "band_std_values": None,
            "scaled_min_values": None,
            "scaled_max_values": None,
            "scaled_mean_values": None,
            "scaled_std_values": None,
        }
    }

    path_save = Path(os.path.split(path)[0])
    normstats_json_path = os.path.abspath(path_save / ".." / stats_file_name)

    if not os.path.exists(normstats_json_path):
        normstats = dummy_stats
        with open(normstats_json_path, "w", encoding="utf-8") as f:
            json.dump(normstats, f, ensure_ascii=False, indent=4)
    else:
        with open(normstats_json_path) as f:
            normstats = json.load(f)

    emd_path = Path(os.path.abspath(path_save / "esri_model_definition.emd"))
    with open(emd_path) as f:
        emd_stats = json.load(f)

    if stats_file_name == "esri_normalization_stats_a.json":
        domain_stats = emd_stats.get("AllTilesStats")
    elif stats_file_name == "esri_normalization_stats_b.json":
        domain_stats = emd_stats.get("AllTilesStats")
    else:
        domain_stats = emd_stats.get("AllTilesStats2")

    mini, maxi, mean, std = [], [], [], []
    for i in domain_stats:
        mini.append(i.get("Min"))
        maxi.append(i.get("Max"))
        mean.append(i.get("Mean"))
        std.append(i.get("StdDev"))

    data_stats = [mini, maxi, mean, std]
    data_stats_tensors = [
        torch.tensor(mini),
        torch.tensor(maxi),
        torch.tensor(mean),
        torch.tensor(std),
    ]

    norm_pct_search = f"batch_stats_for_norm_pct_{round(norm_pct*100)}"
    if norm_pct_search in normstats:
        batch_stats = normstats[norm_pct_search]
        for l, s in enumerate(batch_stats):
            if l == len(data_stats):
                break
            batch_stats[s] = data_stats[l]

        for s in batch_stats:
            if batch_stats[s] is not None:
                batch_stats[s] = torch.tensor(batch_stats[s])
    else:
        batch_stats = {
            "band_min_values": None,
            "band_max_values": None,
            "band_mean_values": None,
            "band_std_values": None,
            "scaled_min_values": None,
            "scaled_max_values": None,
            "scaled_mean_values": None,
            "scaled_std_values": None,
        }
        for l, s in enumerate(batch_stats):
            if l == len(data_stats):
                break
            batch_stats[s] = data_stats[l]

        for s in batch_stats:
            if batch_stats[s] is not None:
                batch_stats[s] = torch.Tensor(batch_stats[s])
        normstats[norm_pct_search] = dict(batch_stats)
        for s in normstats[norm_pct_search]:
            if normstats[norm_pct_search][s] is not None:
                normstats[norm_pct_search][s] = normstats[norm_pct_search][s].tolist()
        with open(normstats_json_path, "w", encoding="utf-8") as f:
            json.dump(normstats, f, ensure_ascii=False, indent=4)

    return batch_stats


def prepare_data_wnetcgan(path, norm_pct, val_split_pct, seed, databunch_kwargs):
    path_a = path / "train_A_C" / "images"
    path_c = path / "train_A_C" / "images2"
    path_b = path / "train_B" / "images"

    img_list_a = ArcGISImageList.from_folder(path_a)
    img_list_b = ArcGISImageList.from_folder(path_b)
    img_list_c = ArcGISImageList.from_folder(path_c)

    batch_stats_a = _batch_stats_json(
        path_a, img_list_a, norm_pct, stats_file_name="esri_normalization_stats_a.json"
    )
    batch_stats_b = _batch_stats_json(
        path_b, img_list_b, norm_pct, stats_file_name="esri_normalization_stats_b.json"
    )
    batch_stats_c = _batch_stats_json(
        path_c, img_list_c, norm_pct, stats_file_name="esri_normalization_stats_c.json"
    )

    data = (
        ImageTupleListMS2.from_folders(
            path,
            path_a,
            path_b,
            path_c,
            batch_stats_a,
            batch_stats_b,
            batch_stats_c,
        )
        .split_by_rand_pct(val_split_pct, seed=seed)
        .label_empty()
        .databunch(**databunch_kwargs)
    )

    data._band_min_values = batch_stats_a["band_min_values"]
    data._band_max_values = batch_stats_a["band_max_values"]
    data._band_mean_values = batch_stats_a["band_mean_values"]
    data._band_std_values = batch_stats_a["band_std_values"]
    data._scaled_min_values = batch_stats_a["scaled_min_values"]
    data._scaled_max_values = batch_stats_a["scaled_max_values"]
    data._scaled_mean_values = batch_stats_a["scaled_mean_values"]
    data._scaled_std_values = batch_stats_a["scaled_std_values"]

    data._band_min_values_b = batch_stats_b["band_min_values"]
    data._band_max_values_b = batch_stats_b["band_max_values"]
    data._band_mean_values_b = batch_stats_b["band_mean_values"]
    data._band_std_values_b = batch_stats_b["band_std_values"]
    data._scaled_min_values_b = batch_stats_b["scaled_min_values"]
    data._scaled_max_values_b = batch_stats_b["scaled_max_values"]
    data._scaled_mean_values_b = batch_stats_b["scaled_mean_values"]
    data._scaled_std_values_b = batch_stats_b["scaled_std_values"]

    data._band_min_values_c = batch_stats_c["band_min_values"]
    data._band_max_values_c = batch_stats_c["band_max_values"]
    data._band_mean_values_c = batch_stats_c["band_mean_values"]
    data._band_std_values_c = batch_stats_c["band_std_values"]
    data._scaled_min_values_c = batch_stats_c["scaled_min_values"]
    data._scaled_max_values_c = batch_stats_c["scaled_max_values"]
    data._scaled_mean_values_c = batch_stats_c["scaled_mean_values"]
    data._scaled_std_values_c = batch_stats_c["scaled_std_values"]

    # add dataset_type
    data._dataset_type = "WNet_cGAN"
    return data


def display_row(axes, display, rgb_bands=None):
    if rgb_bands is None:
        rgb_bands = [0, 1, 2]
    for i, ax in enumerate(axes):
        if type(display[i]) is ArcGISMSImage:
            display[i].show(ax, rgb_bands)
        else:
            ax.imshow(display[i])
        ax.axis("off")


def show_batch_wnet(self, rows=4, **kwargs):
    rgb_bands = kwargs.get("rgb_bands", None)
    fig, axes = plt.subplots(nrows=rows, ncols=3, squeeze=False, figsize=(20, rows * 5))
    top = get_top_padding(title_font_size=16, nrows=rows, imsize=5)
    plt.subplots_adjust(top=top)
    fig.suptitle("Input 1 / Input 2 / Label", fontsize=16)
    img_idxs = [random.randint(0, len(self.train_ds) - 1) for k in range(rows)]
    for idx, im_idx in enumerate(img_idxs):
        self.train_ds.show(im_idx, axes[idx], rgb_bands)


def show_results(self, rows):
    self.learn.model.eval()
    x_batch, y_batch = get_nbatches(
        self._data.valid_dl, math.ceil(rows / self._data.batch_size)
    )

    x_A, x_B, x_C = (
        [x[0] for x in x_batch],
        [x[1] for x in x_batch],
        [x[2] for x in x_batch],
    )

    x_A = torch.cat(x_A)
    x_B = torch.cat(x_B)
    x_C = torch.cat(x_C)

    top = get_top_padding(title_font_size=16, nrows=rows, imsize=5)
    activ = []

    for i in range(0, x_B.shape[0], self._data.batch_size):
        with torch.no_grad():
            preds = self.learn.model(
                x_A[i : i + self._data.batch_size].detach(),
                x_B[i : i + self._data.batch_size].detach(),
                x_C[i : i + self._data.batch_size].detach(),
            )
        if self._data.nband_c == 1:
            all_batchs = []
            for batc in range(preds[0].shape[0]):
                first_pred = preds[0][batc, 1, :, :]
                real_preds = np.array(
                    [first_pred.clone().cpu().numpy()] * self._data.n_channel
                )
                all_batchs.append(real_preds)
            preds_ = torch.tensor(all_batchs).cuda()
            activ.append(preds_)
        else:
            activ.append(preds[0])

    activations = torch.cat(activ)
    if activations.shape[0] == x_A.shape[0]:
        x_A = 0.5 + x_A.cpu() / 2
        x_B = 0.5 + x_B.cpu() / 2
        x_C = 0.5 + x_C.cpu() / 2
        activations = 0.5 + activations.cpu() / 2

        rows = min(rows, x_A.shape[0])
        fig, axs = plt.subplots(
            nrows=rows, ncols=4, figsize=(4 * 5, rows * 5), squeeze=False
        )
        plt.subplots_adjust(top=top)
        fig.suptitle("Input 1 / Input 2 / Label / Predictions")
        for r in range(rows):
            if self._data._is_multispectral:
                display_row(
                    axs[r],
                    (
                        ArcGISMSImage(x_A[r]),
                        ArcGISMSImage(x_B[r]),
                        ArcGISMSImage(x_C[r]),
                        ArcGISMSImage(activations[r]),
                    ),
                )
            else:
                display_row(
                    axs[r],
                    (
                        image2np(x_A[r]),
                        image2np(x_B[r]),
                        image2np(x_C[r]),
                        image2np(activations[r]),
                    ),
                )



# Copyright 2020 The MMSegmentation Authors.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# model settings
norm_cfg = dict(type="BN", requires_grad=True)
model = dict(
    type="EncoderDecoder",
    pretrained="open-mmlab://contrib/mobilenet_v3_large",
    backbone=dict(
        type="MobileNetV3", arch="large", out_indices=(1, 3, 16), norm_cfg=norm_cfg
    ),
    decode_head=dict(
        type="LRASPPHead",
        in_channels=(16, 24, 960),
        in_index=(0, 1, 2),
        channels=128,
        input_transform="multiple_select",
        dropout_ratio=0.1,
        num_classes=19,
        norm_cfg=norm_cfg,
        act_cfg=dict(type="ReLU"),
        align_corners=False,
        loss_decode=dict(type="CrossEntropyLoss", use_sigmoid=False, loss_weight=1.0),
    ),
    # model training and testing settings
    train_cfg=dict(),
    test_cfg=dict(mode="whole"),
)

checkpoint = "https://download.openmmlab.com/mmsegmentation/v0.5/mobilenet_v3/lraspp_m-v3-d8_512x1024_320k_cityscapes/lraspp_m-v3-d8_512x1024_320k_cityscapes_20201224_220337-cfe8fb07.pth"



from dataclasses import dataclass, asdict, field
from typing import Dict, Union, Optional, Any, ClassVar


@dataclass(frozen=True)
class TimeInstant:
    """
    Data class that holds the Instant Time configuration

    =====================      ====================================================================
    **Parameter**               **Description**
    ---------------------      --------------------------------------------------------------------
    time_field                 String. Time field name.
    =====================      ====================================================================

    =====================      ====================================================================
    **Optional Argument**      **Description**
    ---------------------      --------------------------------------------------------------------
    date_format                String. If the field does not contain epoch values,
                               a date format can be defined for the time field.
    =====================      ====================================================================

    :return: boolean `True` if the operation is a success

    .. code-block:: python

        # Usage Example

        time = TimeInstant(time_field="time_field")

    """

    time_field: str
    date_format: str = None


@dataclass(frozen=True)
class TimeInterval:
    """
    Data class that holds the Interval Time configuration

    =====================     ====================================================================
    **Parameter**              **Description**
    ---------------------     --------------------------------------------------------------------
    interval_start_field      String. Start-time field name for the time interval.
    ---------------------     --------------------------------------------------------------------
    interval_end_field        String. End-time field name for the time interval.
    =====================     ====================================================================

    =====================     ====================================================================
    **Optional Argument**     **Description**
    ---------------------     --------------------------------------------------------------------
    date_format               String. If the field does not contain epoch values,
                              a date format can be defined for the time field.
    =====================     ====================================================================

    :return: boolean `True` if the operation is a success

    .. code-block:: python

        # Usage Example

        time = TimeInterval(
            interval_start_field="start_field",
            interval_end_field="end_field"
        )

    """

    interval_start_field: str
    interval_end_field: str
    date_format: str = None


_START_TIME_TAG = "START_TIME"
_END_TIME_TAG = "END_TIME"


class _HasTime:
    # ---> Inheriting classes MUST declare the following Optional fields. <---
    # These are commented variables because of a limitation in dataclass where base class properties get ordered before
    # the derived class properties. Since these are optional properties, the init will order them before the non-optional
    # derived class properties leading to errors like - TypeError: non-default argument 'rss_url' follows default argument
    #
    # time: Optional[Union[TimeInstant, TimeInterval]] = None

    def set_time_config(self, time: Union[TimeInstant, TimeInterval]) -> bool:
        """
        Configures the time property for a feed

        ==============          ====================================================================
        **Parameter**            **Description**
        ---------------         --------------------------------------------------------------------
        time                    [:class:`~arcgis.realtime.velocity.feeds.TimeInstant`, :class:`~arcgis.realtime.velocity.feeds.TimeInterval`].
                                Time object used to configure the feed.
        ===============         ====================================================================

        :return: boolean `True` if the operation is a success

        .. code-block:: python

            # Usage Example
            time = TimeInterval(
                interval_start_field="start_field",
                interval_end_field="end_field"
            )
            feed.set_time_config(time=time)

        """

        if isinstance(time, TimeInstant):
            if time.date_format is not None and self.data_format is not None:
                self.data_format.date_format = time.date_format

            is_success = False
            for field in self._fields["attributes"]:
                if (
                    field["name"] == time.time_field
                    or field["toField"] == time.time_field
                ):
                    field["tags"] = [_START_TIME_TAG]

                    self._fields["time"] = {"timeType": "Instant"}
                    is_success = True
                elif _START_TIME_TAG in field["tags"]:
                    field["tags"].clear()

            if not is_success:
                raise ValueError(f"invalid time_field: '{time.time_field}'")
            else:
                return True

        elif isinstance(time, TimeInterval):
            if time.date_format is not None and self.data_format is not None:
                self.data_format.date_format = time.date_format

            is_success_1 = False
            is_success_2 = False
            for field in self._fields["attributes"]:
                if (
                    field["name"] == time.interval_start_field
                    or field["toField"] == time.interval_start_field
                ):
                    field["tags"] = [_START_TIME_TAG]
                    is_success_1 = True
                elif (
                    field["name"] == time.interval_end_field
                    or field["toField"] == time.interval_end_field
                ):
                    field["tags"] = [_END_TIME_TAG]
                    is_success_2 = True
                elif any(
                    elem in [_START_TIME_TAG, _END_TIME_TAG] for elem in field["tags"]
                ):
                    # if a user had previously assigned _START_TIME_TAG or _END_TIME_TAG tags to other fields
                    # it should be cleared
                    field["tags"].clear()

            if not is_success_1:
                raise ValueError(
                    f"invalid interval_start_field: '{time.interval_start_field}'"
                )
            elif not is_success_2:
                raise ValueError(
                    f"invalid interval_end_field: '{time.interval_end_field}'"
                )
            else:
                self._fields["time"] = {"timeType": "Interval"}

                return True

    def reset_time_config(self) -> bool:
        """
        Removes any previously set time configuration from the schema.

        :return: boolean `True` if the operation is a success. `False` if a previously set time configuration was not found.
        """

        is_success = False
        # remove time tags from all field attributes
        for field in self._fields["attributes"]:
            if any(elem in [_START_TIME_TAG, _END_TIME_TAG] for elem in field["tags"]):
                field["tags"].clear()
                is_success = True

        if is_success:
            # remove the "time": {"timeType": "Instant/Interval"} property from _fields
            self._fields.pop("time", None)

        return is_success



# Copyright 2018-2019 Open-MMLab.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Currently only batch_size 1 for inference mode is supported. Found batch_size 2. error is comming so we don't support for now
# detr_r50_8x2_150e_coco.py, box AP=40.1
model = dict(
    type="DETR",
    pretrained="torchvision://resnet50",
    backbone=dict(
        type="ResNet",
        depth=50,
        num_stages=4,
        out_indices=(3,),
        frozen_stages=1,
        norm_cfg=dict(type="BN", requires_grad=False),
        norm_eval=True,
        style="pytorch",
    ),
    bbox_head=dict(
        type="TransformerHead",
        num_classes=80,
        in_channels=2048,
        num_fcs=2,
        transformer=dict(
            type="Transformer",
            embed_dims=256,
            num_heads=8,
            num_encoder_layers=6,
            num_decoder_layers=6,
            feedforward_channels=2048,
            dropout=0.1,
            act_cfg=dict(type="ReLU", inplace=True),
            norm_cfg=dict(type="LN"),
            num_fcs=2,
            pre_norm=False,
            return_intermediate_dec=True,
        ),
        positional_encoding=dict(
            type="SinePositionalEncoding", num_feats=128, normalize=True
        ),
        loss_cls=dict(
            type="CrossEntropyLoss",
            bg_cls_weight=0.1,
            use_sigmoid=False,
            loss_weight=1.0,
            class_weight=1.0,
        ),
        loss_bbox=dict(type="L1Loss", loss_weight=5.0),
        loss_iou=dict(type="GIoULoss", loss_weight=2.0),
    ),
    # training and testing settings
    train_cfg=dict(
        assigner=dict(
            type="HungarianAssigner",
            cls_cost=dict(type="ClassificationCost", weight=1.0),
            reg_cost=dict(type="BBoxL1Cost", weight=5.0),
            iou_cost=dict(type="IoUCost", iou_mode="giou", weight=2.0),
        )
    ),
    test_cfg=dict(max_per_img=100),
)

checkpoint = "http://download.openmmlab.com/mmdetection/v2.0/detr/detr_r50_8x2_150e_coco/detr_r50_8x2_150e_coco_20201130_194835-2c4b8974.pth"



""" Utility functions internally used by the store.
"""

from arcgis.features import FeatureSet
from ... import workforce


def _should_use_async_apply_edits(feature_layer):
    return (
        "advancedEditingCapabilities" in feature_layer.properties
        and "supportsAsyncApplyEdits"
        in feature_layer.properties["advancedEditingCapabilities"]
        and feature_layer.properties["advancedEditingCapabilities"][
            "supportsAsyncApplyEdits"
        ]
    )


def add_features(feature_layer, features, use_global_ids=False):
    """Adds features to the feature_layer.  The input features will be updated upon successful
    adding on the server, such that they contain the server-assigned object_ids and global_ids.
    :param feature_layer: An arcgis.features.FeatureLayer.
    :param features: list of arcgis.features.Features.
    :param use_global_ids: use global ids or not
    :returns: The added features.
    :raises ServerError: Indicates that the server rejected the new features.
    """
    if features:
        feature_set = FeatureSet(features)
        response = feature_layer.edit_features(
            adds=feature_set,
            use_global_ids=use_global_ids,
            future=_should_use_async_apply_edits(feature_layer),
        )
        if _should_use_async_apply_edits(feature_layer):
            add_results = response.result()[0]["addResults"]
        else:
            add_results = response["addResults"]
        errors = [result["error"] for result in add_results if not result["success"]]
        if errors:
            raise workforce.ServerError(errors)
        for feature, add_results in zip(features, add_results):
            feature.attributes[feature_layer.properties("objectIdField")] = add_results[
                "objectId"
            ]
            feature.attributes[feature_layer.properties("globalIdField")] = add_results[
                "globalId"
            ]
    return features


def update_features(feature_layer, features):
    """Updates features in a feature_layer.
    :param feature_layer: An arcgis.features.FeatureLayer.
    :param features: list of arcgis.features.Features.  Each feature must have an object id.
    :raises ServerError: Indicates that the server rejected the updates.
    """
    if features:
        response = feature_layer.edit_features(
            updates=FeatureSet(features),
            future=_should_use_async_apply_edits(feature_layer),
        )
        if _should_use_async_apply_edits(feature_layer):
            update_results = response.result()[0]["updateResults"]
        else:
            update_results = response["updateResults"]
        errors = [result["error"] for result in update_results if not result["success"]]
        if errors:
            raise workforce.ServerError(errors)
    return features


def remove_features(feature_layer, features):
    """Removes features from a feature_layer.
    :param feature_layer: An arcgis.features.FeatureLayer.
    :param features: list of arcgis.features.Features.  Each feature must have an object id.
    :raises ServerError: Indicates that the server rejected the removals.
    """
    if features:
        object_id_attr = feature_layer.properties["objectIdField"]
        object_ids = ",".join(
            [str(feature.attributes[object_id_attr]) for feature in features]
        )
        response = feature_layer.edit_features(
            deletes=object_ids, future=_should_use_async_apply_edits(feature_layer)
        )
        if _should_use_async_apply_edits(feature_layer):
            delete_results = response.result()[0]["deleteResults"]
        else:
            delete_results = response["deleteResults"]
        errors = [result["error"] for result in delete_results if not result["success"]]
        if errors:
            raise workforce.ServerError(errors)


def validate(validate_fn, **kwargs):
    """Runs a Model validation routine, and raises the first ValidationError if any are returned.
    :param validate_fn: A function that takes no positional arguments, and returns a list of
    ValidationErrors.  Any **kwargs given to this function will be passed to validate_fn.
    :raises ValidationError: Indicates that the validation_fn returned a ValidationError.
    """
    validation_failures = validate_fn(**kwargs)
    if validation_failures:
        raise validation_failures[0]



# Copyright 2018-2019 Open-MMLab.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# error in training loop
# retinanet_r50_nasfpn_crop640_50e_coco.py, box AP= 40.5

_base_ = "./_base_/models/retinanet_r50_fpn.py"
cudnn_benchmark = True
# model settings
norm_cfg = dict(type="BN", requires_grad=True)
model = dict(
    type="RetinaNet",
    pretrained="torchvision://resnet50",
    backbone=dict(
        type="ResNet",
        depth=50,
        num_stages=4,
        out_indices=(0, 1, 2, 3),
        frozen_stages=1,
        norm_cfg=norm_cfg,
        norm_eval=False,
        style="pytorch",
    ),
    neck=dict(type="NASFPN", stack_times=7, norm_cfg=norm_cfg),
    bbox_head=dict(type="RetinaSepBNHead", num_ins=5, norm_cfg=norm_cfg),
    # training and testing settings
    train_cfg=dict(assigner=dict(neg_iou_thr=0.5)),
)

checkpoint = "http://download.openmmlab.com/mmdetection/v2.0/nas_fpn/retinanet_r50_nasfpn_crop640_50e_coco/retinanet_r50_nasfpn_crop640_50e_coco-0ad1f644.pth"



"""
The ``arcgis.raster`` module contains classes and analysis functions for working with raster data and
imagery layers.

Raster data is made up of a grid of cells, where each cell or pixel can have a value. It is useful
for storing data that varies continuously, as in a satellite image, a surface of chemical concentrations, or
an elevation surface.

Use :meth:`~arcgis.raster.analytics.is_supported` to check if raster analysis is supported in your GIS.
"""

from ._layer import (
    ImageryLayer,
    ImageryLayerCacheManager,
    RasterManager,
    ImageryTileManager,
    RasterCatalogItem,
    Raster,
    _ImageServerRaster,
    _ArcpyRaster,
    RasterCollection,
)
from .analytics import *
from .functions import *
from .orthomapping import *
from .utils import *

# from .charts import *
from ._RasterInfo import RasterInfo



from arcgis.auth.tools import LazyLoader
import os
import functools
import urllib.parse
from arcgis.gis import GIS

logging = LazyLoader("logging")
_isd = LazyLoader("arcgis._impl.common._isd")
layer = LazyLoader("arcgis.features.layer")
geocoding = LazyLoader("arcgis.geocoding")
gptool = LazyLoader("arcgis.geoprocessing._tool")
geommodule = LazyLoader("arcgis._impl.tools")
network_dataset = LazyLoader("arcgis.network")
_gis = LazyLoader("arcgis.gis")
mapping = LazyLoader("arcgis.mapping")
raster = LazyLoader("arcgis.raster")
schematics = LazyLoader("arcgis.schematics")

_log = logging.getLogger()


@functools.lru_cache(maxsize=250)
def _create_service(url: str, layer_type: str, gis: GIS, name: str = None):
    has_layer = False
    digit = None
    if name:
        url = url.replace(name, urllib.parse.quote(name))
    if os.path.basename(url).isdigit():
        digit = os.path.basename(url)
        url = os.path.dirname(url)
        has_layer = True
    if layer_type.lower() == "mapserver":
        if has_layer:
            return mapping.MapServiceLayer(url=f"{url}/{digit}", gis=gis)
        else:
            return mapping.MapImageLayer(url=url, gis=gis)
    elif layer_type.lower() == "featureserver":
        if has_layer:
            return layer.FeatureLayer(url=url, gis=gis)
        else:
            return layer.FeatureLayerCollection(url=url, gis=gis)
    elif layer_type.lower() == "imageserver":
        return raster.ImageryLayer(url=url, gis=gis)
    elif layer_type.lower() == "gpserver":
        return gptool.Toolbox(url=url, gis=gis)
    elif layer_type.lower() == "geometryserver":
        return geommodule.GeometryService(url=url, gis=gis)
    elif layer_type.lower() == "geocodeserver":
        return geocoding.Geocoder(location=url, gis=gis)
    elif layer_type.lower() == "naserver":
        return network_dataset.NetworkDataset(url=url, gis=gis)
    elif layer_type.lower() == "vectortileserver":
        return mapping.VectorTileLayer(url=url, gis=gis)
    elif layer_type.lower() == "sceneserver":
        return mapping._scenelyrs._lyrs.SceneLayer(url=url, gis=gis)
    else:
        return _gis.Layer(url=url, gis=gis)

    return None


###########################################################################
class AGOLServicesDirectory:
    """
    The ArcGIS Online Services Directory displays the hosted services for
    a site.

    ==================     ====================================================================
    **Parameter**           **Description**
    ------------------     --------------------------------------------------------------------
    url                    Required String. The url string to the ArcGIS Online Server
    ------------------     --------------------------------------------------------------------
    gis                    Required GIS. The connection to ArcGIS Online.
    ==================     ====================================================================

    """

    _gis = None
    _url = None
    _properties = None

    # ---------------------------------------------------------------------
    def __init__(self, url: str, gis: GIS):
        """initializer"""
        self._gis = gis
        self._url = url

    # ---------------------------------------------------------------------
    def __str__(self):
        return f"< AGOLServicesDirectory @ {self._url} >"

    # ---------------------------------------------------------------------
    def __repr__(self):
        return f"< AGOLServicesDirectory @ {self._url} >"

    # ---------------------------------------------------------------------
    @functools.lru_cache(maxsize=255)
    def _org_id(self, gis: GIS) -> str:
        return gis.properties.id

    # ---------------------------------------------------------------------
    @property
    def properties(self) -> _isd.InsensitiveDict:
        """
        Returns the server's properties

        :returns: InsensitiveDict
        """
        resp = self._gis._con.get(self._url, {"f": "json"})
        return _isd.InsensitiveDict(resp)

    # ---------------------------------------------------------------------
    @property
    def folders(self) -> list:
        """
        Returns a list of folder names

        :returns: List
        """
        return []

    # ---------------------------------------------------------------------
    @property
    def services(self) -> list:
        """returns a list of services hosted on ArcGIS Online Server"""
        services = []

        if "services" in self.properties:
            for s in self.properties["services"]:
                try:
                    services.append(
                        _create_service(
                            url=s["url"],
                            layer_type=s["type"],
                            gis=self._gis,
                            name=s.get("name", None),
                        )
                    )
                except Exception as e:
                    msg = "URL: %s is throwing error:  %s" % (s["url"], e)
                    _log.warning(msg)
                    _log.warning("Could not load service: %s" % s["url"])
        return services



""" Defines store functions for working with Dispatchers.
"""

import math
from ... import workforce
from .utils import add_features, update_features, remove_features, validate


def get_dispatcher(project, object_id=None, global_id=None, user_id=None):
    """Gets the identified Dispatcher.  Exactly one form of identification should be provided.
    :param project:
    :param object_id: The dispatcher's OBJECTID.
    :param global_id: The dispatcher's GlobalID.
    :param user_id: The dispatcher's named user user_id.
    :returns: Dispatcher
    """
    if object_id:
        where = "{} = {}".format(project._dispatcher_schema.object_id, object_id)
    elif global_id:
        where = "{} = '{}'".format(project._dispatcher_schema.global_id, global_id)
    elif user_id:
        where = "{} = '{}'".format(project._dispatcher_schema.user_id, user_id)
    else:
        where = "1=0"
    dispatchers = query_dispatchers(project, where)
    return dispatchers[0] if dispatchers else None


def get_dispatchers(project):
    """Gets all Dispatchers in the project.
    :param project:
    :returns: list of Dispatchers
    """
    return query_dispatchers(project, "1=1")


def query_dispatchers(project, where):
    """Executes a query against the dispatchers feature layer.
    :param project: The project in which to query dispatchers.
    :param where: An ArcGIS where clause.
    :returns: list of Dispatchers
    """
    features = project.dispatchers_layer.query(where, return_all_records=True).features
    return [workforce.Dispatcher(project, feature) for feature in features]


def add_dispatcher(project, feature=None, contact_number=None, name=None, user_id=None):
    """
    Adds a new dispatcher to the project
    """
    project._update_cached_objects()
    dispatcher = workforce.Dispatcher(project, feature, contact_number, name, user_id)
    return add_dispatchers(project, [dispatcher])[0]


def update_dispatcher(
    project, dispatcher, contact_number=None, name=None, user_id=None
):
    """
    Updates a dispatcher and submits changes to the server
    """
    project._update_cached_objects()
    if contact_number:
        dispatcher.contact_number = contact_number
    if name:
        dispatcher.name = name
    if user_id and dispatcher.user_id != project.owner_user_id:
        dispatcher.user_id = user_id
    return update_dispatchers(project, [dispatcher])[0]


def add_dispatchers(project, dispatchers):
    """Adds Dispatchers to a project.

    Side effect: Upon successful addition on the server, the object_id and global_id fields of
    each Dispatcher in dispatchers will be updated to the values assigned by the server.

    :param project:
    :param dispatchers: list of Dispatchers
    :raises ValidationError: Indicates that one or more dispatchers failed validation.
    :raises ServerError: Indicates that the server rejected the dispatchers.
    """
    project._update_cached_objects()
    if dispatchers:
        use_global_ids = True
        for dispatcher in dispatchers:
            validate(dispatcher._validate_for_add)

            if dispatcher.global_id is None:
                use_global_ids = False

        features = [dispatcher.feature for dispatcher in dispatchers]
        add_features(project.dispatchers_layer, features, use_global_ids)

    # add dispatcher named users to the project's group.
    max_add_per_call = 25
    for i in range(0, math.ceil(len(dispatchers) / max_add_per_call)):
        project.group.add_users(
            [
                d.user_id
                for d in dispatchers[
                    i * max_add_per_call : (i * max_add_per_call) + max_add_per_call
                ]
            ]
        )
    return dispatchers


def update_dispatchers(project, dispatchers):
    """Updates Dispatchers.
    :param project:
    :param dispatchers: list of Dispatchers to update
    :raises ValidationError: Indicates that one or more dispatchers failed validation.
    :raises ServerError: Indicates that the server rejected the dispatchers.
    """
    project._update_cached_objects()
    if dispatchers:
        for dispatcher in dispatchers:
            validate(dispatcher._validate_for_update)
        features = [dispatcher.feature for dispatcher in dispatchers]
        update_features(project.dispatchers_layer, features)
    return dispatchers


def delete_dispatchers(project, dispatchers):
    """Removes Dispatchers from the project.
    :param project:
    :param dispatchers: list of Dispatchers
    :raises ValidationError: Indicates that one or more dispatchers failed validation.
    :raises ServerError: Indicates that the server rejected the removal.
    """
    project._update_cached_objects()
    if dispatchers:
        for dispatcher in dispatchers:
            validate(dispatcher._validate_for_remove)
        features = [dispatcher.feature for dispatcher in dispatchers]
        remove_features(project.dispatchers_layer, features)

        # Remove dispatcher named users from the project's group, unless they are also workers.
        user_ids = [dispatcher.user_id for dispatcher in dispatchers]
        where = "{} in ({})".format(
            project._worker_schema.user_id,
            ",".join(["'{}'".format(user_id) for user_id in user_ids]),
        )
        workers = workforce._store.query_workers(project, where)
        for worker in workers:
            user_ids.remove(worker.user_id)
        if user_ids:
            project.group.remove_users(user_ids)



from __future__ import annotations
from typing import Any, Optional, Union

from ._util import (
    _to_datetime,
    _datetime2ole,
    _time_filter,
    _linear_regression,
    _harmonic_regression,
)
import datetime as _datetime
import copy as _copy
import logging as _logging
from arcgis.geometry import Point, Polygon

_LOGGER = _logging.getLogger(__name__)


try:
    import numpy as _np
    import matplotlib.pyplot as _plt
    from matplotlib.pyplot import cm as _cm

except:
    raise


def temporal_profile(
    raster,
    points: list[Point] = [],
    time_field: Optional[str] = None,
    variables: list[str] = [],
    bands: list[int] = [0],
    time_extent: Optional[list[_datetime.datetime]] = None,
    dimension: Optional[list[str]] = None,
    dimension_values: list[int] = [],
    show_values: bool = False,
    trend_type: Optional[str] = None,
    trend_order: Optional[int] = None,
    plot_properties: dict[str, Any] = {},
):
    """
    A temporal profile serves as a basic analysis tool for imagery data in a time series.
    Visualizing change over time with the temporal profile allows trends to be displayed
    and compared with variables, bands, or values from other dimensions simultaneously.

    Using the functionality in temporal profile charts, you can perform trend analysis, gain insight into
    multidimensional raster data at given locations, and plot values that are changing over time
    in the form of a line graph.

    Temporal profile charts can be used in various scientific applications involving time series
    analysis of raster data, and the graphical output of results can be used directly as
    input for strategy management and decision making.

    The x-axis of the temporal profile displays the time in continuous time intervals. The time field is
    obtained from the timeInfo of the image service.

    The y-axis of the temporal profile displays the variable value.


    ====================================     ====================================================================
    **Parameter**                             **Description**
    ------------------------------------     --------------------------------------------------------------------
    raster                                   Required Imagery Layer object.
    ------------------------------------     --------------------------------------------------------------------
    points                                   Required list of point Geometry objects.
    ------------------------------------     --------------------------------------------------------------------
    time_field                               Required string. The time field that will be used for plotting
                                             temporal profile.

                                             If not specified the time field is obtained from the timeInfo of
                                             the image service.
    ------------------------------------     --------------------------------------------------------------------
    variables                                Required list of variable names.
                                             For non multidimensional data, the variable would be name of the Sensor.
                                             To plot the graph against all sensors specify - "ALL_SENSORS"
    ------------------------------------     --------------------------------------------------------------------
    bands                                    Optional list of band indices. By default takes the
                                             first band (band index - 0).
                                             For a multiband data, you can compare the time change of different
                                             bands over different locations.
    ------------------------------------     --------------------------------------------------------------------
    time_extent                              Optional list of date time object. This represents the time extent
    ------------------------------------     --------------------------------------------------------------------
    dimension                                Optional list of dimension names. This option works specifically on
                                             multidimensional data containing a time dimension and other dimensions.

                                             The temporal profile is created based on the specific values in other
                                             dimensions, such as depth at the corresponding time value. For example,
                                             soil moisture data usually includes both a time dimension and vertical
                                             dimension below the earth's surface, resulting in a temporal profile
                                             at 0.1, 0.2, and 0.3 meters below the ground.
    ------------------------------------     --------------------------------------------------------------------
    dimension_values                         Optional list of dimension values. This parameter can be used to specify
                                             the values of dimension parameter other than the time dimension (dimension
                                             name specified using dimension parameter)
    ------------------------------------     --------------------------------------------------------------------
    show_values                              Optional bool. Default False.
                                             Set this parameter to True to display the values at each point in the line graph.
    ------------------------------------     --------------------------------------------------------------------
    trend_type                               Optional string. Default None.
                                             Set the trend_type parameter eith with linear or harmonic to draw the trend line
                                             linear : Fits the pixel values for a variable along a linear trend line.
                                             harmonic : Fits the pixel values for a variable along a harmonic trend line.
    ------------------------------------     --------------------------------------------------------------------
    trend_order                              optional number. The frequency number to use in the trend fitting.
                                             This parameter specifies the frequency of cycles in a year.
                                             The default value is 1, or one harmonic cycle per year.

                                             This parameter is only included in the trend analysis for a harmonic regression.
    ------------------------------------     --------------------------------------------------------------------
    plot_properties                          Optional dict. This parameter can be used to set the figure
                                             properties. These are the matplotlib.pyplot.figure() parameters and values
                                             specified in dict format.

                                             eg: {"figsize":(15,15)}
    ====================================     ====================================================================

    :return:
        None

    """

    t1 = []

    if not isinstance(variables, list):
        variables = [variables]
    if not isinstance(points, list):
        points = [points]
    if not isinstance(bands, list):
        bands = [bands]

    if time_field is None:
        try:
            x_var = raster.properties.timeInfo["startTimeField"]
        except:
            raise RuntimeError("Specify time_field to plot the temporal profile.")
    else:
        x_var = time_field

    draw_trend = False
    if trend_type is not None:
        if trend_type.lower() in ["linear", "harmonic"]:
            draw_trend = True
            if trend_type.lower() == "harmonic":
                if trend_order is None:
                    trend_order = 1
                    _LOGGER.warning(
                        "Invalid Argument - trend order is None. Setting trend order as 1 to plot the trend line"
                    )
                elif trend_order < 1:
                    trend_order = 1
                    _LOGGER.warning(
                        "Invalid Argument - trend order is less than 1. Setting trend order as 1 to plot the trend line"
                    )
                elif trend_order > 3:
                    trend_order = 3
                    _LOGGER.warning(
                        "Invalid Argument - trend order is greater than 3. Setting trend order as 3 to plot the trend line"
                    )
        else:
            _LOGGER.warning(
                "Trend line cannot be drawn. Trend type should be one of the following: linear, harmonic "
            )

    if (
        "hasMultidimensions" in raster.properties
        and raster.properties["hasMultidimensions"] == True
    ):
        mosaic_rule = {
            "mosaicMethod": "esriMosaicAttribute",
            "ascending": False,
            "sortField": x_var,
            "multidimensionalDefinition": [
                {"variableName": "", "dimensionName": x_var}
            ],
        }
        if time_extent is not None:
            if isinstance(time_extent, _datetime.datetime):
                time_extent = [int(time_extent.timestamp() * 1000)]
            elif isinstance(time_extent, list):
                if isinstance(time_extent[0], _datetime.datetime) and isinstance(
                    time_extent[1], _datetime.datetime
                ):
                    time_extent = [
                        int(time_extent[0].timestamp() * 1000),
                        int(time_extent[1].timestamp() * 1000),
                    ]
            for index, each_elem in enumerate(
                mosaic_rule["multidimensionalDefinition"]
            ):
                if (
                    mosaic_rule["multidimensionalDefinition"][index]["dimensionName"]
                    == x_var
                ):
                    mosaic_rule["multidimensionalDefinition"][index]["values"] = [
                        time_extent
                    ]

        if dimension is not None and dimension_values is not None:
            if not isinstance(dimension_values, list):
                dimension_values = [dimension_values]
            mosaic_rule["multidimensionalDefinition"].append(
                {
                    "variableName": "",
                    "dimensionName": dimension,
                    "values": dimension_values,
                    "isSlice": True,
                }
            )

        num_lines = len(dimension_values) * len(variables) * len(points) * len(bands)
        y = [[] for i in range(0, num_lines)]
        x = [[] for i in range(0, num_lines)]
        # x_var = raster.properties.timeInfo['startTimeField']

        n_dim_val = 1 if len(dimension_values) == 0 else len(dimension_values)
        n_bands = 1 if len(bands) == 0 else len(bands)
        n_colors = n_dim_val * len(variables) * len(points) * n_bands
        if trend_type is not None:
            n_colors = n_colors * 2

        if len(variables) == 1:
            variable_unit = None
            for ele in raster.multidimensional_info["multidimensionalInfo"][
                "variables"
            ]:
                if ele["name"] == variables[0]:
                    if "unit" in ele.keys():
                        variable_unit = ele["unit"]
                        break
        res = []
        t1 = []
        d1 = []
        xx = []
        yy = []
        for band in bands:
            for index, point in enumerate(points):
                for variable in variables:
                    for md_def in mosaic_rule["multidimensionalDefinition"]:
                        md_def["variableName"] = variable

                    res = raster.get_samples(
                        geometry=point,
                        return_first_value_only=False,
                        out_fields="*",
                        mosaic_rule=mosaic_rule,
                    )

                    if dimension_values != []:
                        for dim_value in dimension_values:
                            for res_ele in res:
                                if res_ele["attributes"][dimension] == dim_value:
                                    yy.append(res_ele["values"][band])
                                    xx.append(
                                        _to_datetime(res_ele["attributes"][x_var])
                                    )
                            d1.append(
                                {"yy": yy, "xx": xx, "dimension_value": dim_value}
                            )
                            yy = []
                            xx = []

                    else:
                        for ele in res:
                            y.append(ele["values"][band])
                            x.append(_to_datetime(ele["attributes"][x_var]))

                    if dimension_values == []:
                        t1.append(
                            {
                                "y": y,
                                "x": x,
                                "point": index,
                                "variable": variable,
                                "band": band,
                            }
                        )
                    else:
                        for ele in d1:
                            t1.append(
                                {
                                    "y": ele["yy"],
                                    "x": ele["xx"],
                                    "point": index,
                                    "variable": variable,
                                    "dimension_value": ele["dimension_value"],
                                    "dimension": dimension,
                                    "band": band,
                                }
                            )
                    x = []
                    y = []
                    d1 = []

        if plot_properties is None:
            plot_properties = {}
        if len(plot_properties) == 0 or (
            len(plot_properties) > 0 and "figsize" not in plot_properties.keys()
        ):
            plot_properties.update({"figsize": (15, 15)})
        if plot_properties is not None and isinstance(plot_properties, dict):
            # {"figsize":(20,10),"dpi":100,"facecolor":"yellow","edgecolor":"blue","linewidth":10.0,"frameon":False}
            _plt.figure(**plot_properties)
        _plt.xlabel(x_var)
        if len(variables) == 1:
            if variable_unit is not None and variable_unit.lower() != "none":
                _plt.ylabel(variables[0] + " (in " + variable_unit + ")")
            else:
                _plt.ylabel(variables[0])
        else:
            _plt.ylabel("Values")

        title_string = "Change in"
        for ele in variables:
            title_string = title_string + " " + str(ele + ",")
        title_string = title_string + " over " + x_var
        if dimension is not None and dimension_values is not None:
            title_string = title_string + "," " at " + str(dimension) + " =  " + str(
                dimension_values
            )
        _plt.title(title_string)

        color = iter(_cm.rainbow(_np.linspace(0, 1, n_colors)))
        for i in range(0, len(t1)):
            label_string = (
                "Location " + str(t1[i]["point"]) + "-" + str(t1[i]["variable"])
            )
            if "dimension" in t1[i].keys():
                label_string = (
                    label_string
                    + "-"
                    + t1[i]["dimension"]
                    + "="
                    + str(t1[i]["dimension_value"])
                )
            if "band" in t1[i].keys():
                label_string = label_string + "-" + "band " + str(t1[i]["band"])
            c = next(color)
            _plt.plot(t1[i]["x"], t1[i]["y"], c=c, label=label_string)
            _plt.scatter(t1[i]["x"], t1[i]["y"], c=[c])
            _plt.legend(loc="center left", bbox_to_anchor=(1, 0.5))

            # for i in range(0,len(t1)):
            #    label_string =  "Location "+ str(t1[i]["point"])+"-"+ str(t1[i]["variable"])
            #    if "dimension" in t1[i].keys():
            #        label_string = label_string+"-"+t1[i]["dimension"]+"="+str(t1[i]["dimension_value"])
            #    plt.plot(t1[i]["x"],t1[i]["y"], label=label_string)
            #    plt.scatter(t1[i]["x"],t1[i]["y"])
            #    plt.legend(loc='upper left')
            # plt.gcf().autofmt_xdate()
            # print(t1[i]["x"]," < ", t1[i]["y"])

            if draw_trend:
                c = next(color)
                date_list = []
                for date in t1[i]["x"]:
                    ole_date = _datetime2ole(date)
                    date_list.append(ole_date)
                sample_size = len(date_list)
                if sample_size != len(t1[i]["y"]):
                    print("error")
                if trend_type.lower() == "linear":
                    x_trend, y_trend = _linear_regression(
                        sample_size, date_list, t1[i]["x"], t1[i]["y"]
                    )
                elif trend_type.lower() == "harmonic":
                    x_trend, y_trend = _harmonic_regression(
                        sample_size, date_list, t1[i]["x"], t1[i]["y"], trend_order
                    )
                if x_trend != [] and y_trend != []:
                    label_string = f"Location {str(t1[i]['point'])}-{str(t1[i]['variable'])}-band {str(t1[i]['band'])}-{trend_type.lower()} trend"
                    _plt.plot(
                        x_trend, y_trend, c=c, linestyle="dashed", label=label_string
                    )
                    _plt.legend(loc="center left", bbox_to_anchor=(1, 0.5))

            if show_values:
                for x, y in zip(t1[i]["x"], t1[i]["y"]):
                    label = "{:.2f}".format(y)
                    _plt.annotate(
                        label,  # this is the text
                        (x, y),  # this is the point to label
                        textcoords="offset points",  # how to position the text
                        xytext=(10, 5),  # distance from text to points (x,y)
                        ha="center",
                    )

        _plt.show()
        # plt.legend()
    else:
        num_lines = len(points) * len(bands)
        # y=[[] for i in range(0, num_lines)]
        # x=[[] for i in range(0, num_lines)]
        t2 = []
        t1 = []
        xx = []
        yy = []
        d1 = []

        mosaic_rule = {
            "mosaicMethod": "esriMosaicAttribute",
            "ascending": False,
            "sortField": x_var,
        }
        for index, point in enumerate(points):
            for variable in variables:
                t2 = raster.get_samples(
                    geometry=point,
                    return_first_value_only=False,
                    out_fields="*",
                    mosaic_rule=mosaic_rule,
                )
                # print(t2)
                # x_var = raster.properties.timeInfo['startTimeField']

                for band in bands:
                    for element in t2:
                        if "attributes" in element:
                            if "SensorName" in element["attributes"].keys():
                                if variable.upper() == "ALL_SENSORS":
                                    if (
                                        _time_filter(
                                            time_extent,
                                            _to_datetime(element["attributes"][x_var]),
                                        )
                                        == True
                                    ):
                                        yy.append(element["values"][band])
                                        xx.append(
                                            _to_datetime(element["attributes"][x_var])
                                        )
                                        xx, yy = zip(*sorted(zip(xx, yy)))
                                        xx = list(xx)
                                        yy = list(yy)
                                if element["attributes"]["SensorName"] == variable:
                                    if (
                                        _time_filter(
                                            time_extent,
                                            _to_datetime(element["attributes"][x_var]),
                                        )
                                        == True
                                    ):
                                        yy.append(element["values"][band])
                                        xx.append(
                                            _to_datetime(element["attributes"][x_var])
                                        )
                                        xx, yy = zip(*sorted(zip(xx, yy)))
                                        xx = list(xx)
                                        yy = list(yy)

                    d1.append({"yy": yy, "xx": xx, "band": band})
                    yy = []
                    xx = []
                for ele in d1:
                    t1.append(
                        {
                            "y": ele["yy"],
                            "x": ele["xx"],
                            "point": index,
                            "variable": variable,
                            "band": ele["band"],
                        }
                    )
                d1 = []
        # print(t1)
        if plot_properties is None:
            plot_properties = {}
        if len(plot_properties) == 0 or (
            len(plot_properties) > 0 and "figsize" not in plot_properties.keys()
        ):
            plot_properties.update({"figsize": (15, 15)})
        if plot_properties is not None and isinstance(plot_properties, dict):
            # {"figsize":(20,10),"dpi":100,"facecolor":"yellow","edgecolor":"blue","linewidth":10.0,"frameon":False}
            _plt.figure(**plot_properties)

        _plt.xlabel(x_var)
        _plt.ylabel(variable)
        if len(variables) == 1:
            _plt.ylabel(variables[0])
        else:
            _plt.ylabel("Values")
        title_string = "Change in"
        for ele in variables:
            title_string = title_string + " " + str(ele + ",")
        title_string = title_string + " over " + x_var
        _plt.title(title_string)

        n_bands = 1 if len(bands) == 0 else len(bands)
        n_colors = len(variables) * len(points) * n_bands
        if trend_type is not None:
            n_colors = n_colors * 2

        color = iter(_cm.rainbow(_np.linspace(0, 1, n_colors)))
        for i in range(0, len(t1)):
            label_string = (
                "Location " + str(t1[i]["point"]) + "-" + str(t1[i]["variable"])
            )
            # label_string =  "Location "+ str(t1[i]["point"])+"-"
            if "band" in t1[i].keys():
                label_string = label_string + "-" + "band " + str(t1[i]["band"])
            c = next(color)
            _plt.plot(t1[i]["x"], t1[i]["y"], c=c, label=label_string)
            _plt.scatter(t1[i]["x"], t1[i]["y"], c=[c])
            _plt.legend(loc="center left", bbox_to_anchor=(1, 0.5))

            if draw_trend:
                c = next(color)
                date_list = []
                for date in t1[i]["x"]:
                    ole_date = _datetime2ole(date)
                    date_list.append(ole_date)
                sample_size = len(date_list)
                if sample_size != len(t1[i]["y"]):
                    print("error")
                if trend_type.lower() == "linear":
                    x_trend, y_trend = _linear_regression(
                        sample_size, date_list, t1[i]["x"], t1[i]["y"]
                    )
                elif trend_type.lower() == "harmonic":
                    x_trend, y_trend = _harmonic_regression(
                        sample_size, date_list, t1[i]["x"], t1[i]["y"], trend_order
                    )
                if x_trend != [] and y_trend != []:
                    label_string = f"Location {str(t1[i]['point'])}-{str(t1[i]['variable'])}-band {str(t1[i]['band'])}-{trend_type.lower()} trend"
                    _plt.plot(
                        x_trend, y_trend, c=c, linestyle="dashed", label=label_string
                    )
                    _plt.legend(loc="center left", bbox_to_anchor=(1, 0.5))

            if show_values:
                for x, y in zip(t1[i]["x"], t1[i]["y"]):
                    label = "{:.2f}".format(y)
                    _plt.annotate(
                        label,  # this is the text
                        (x, y),  # this is the point to label
                        textcoords="offset points",  # how to position the text
                        xytext=(10, 5),  # distance from text to points (x,y)
                        ha="center",
                    )
        # _plt.gcf().autofmt_xdate()
        _plt.show()


def plot_histograms(
    raster,
    geometry: Optional[Union[Polygon, dict[str, Any]]] = None,
    pixel_size: Optional[Union[list[float], dict[str, Any]]] = None,
    time: Optional[Union[_datetime.datetime, _datetime.date, str]] = None,
    bands: list[int] = [],
    display_stats: bool = True,
    plot_properties: Optional[dict[str, Any]] = None,
    subplot_properties: Optional[Union[list[dict], dict[str, Any]]] = None,
):
    """
    Image histograms visually summarize the distribution of a continuous numeric variable by measuring 
    the frequency at which certain values appear in the image. The x-axis in the image histogram is a 
    number line that displays the range of image pixel values that has been split into number ranges, 
    or bins. A bar is drawn for each bin, and the width of the bar represents the density number range 
    of the bin; the height of the bar represents the number of pixels that fall into that range. 
    Understanding the distribution of your data is an important step in the data exploration process.

    ``plot_histograms()`` can be used for plotting the band-wise image histogram charts of any Raster object.

    ============================    ====================================================================
    **Parameter**                   **Description**
    ----------------------------    --------------------------------------------------------------------
    geometry                        Optional Polygon or Extent. A geometry that defines the geometry
                                    within which the histogram is computed. The geometry can be an
                                    envelope or a polygon. If not provided, then the full extent of the 
                                    raster will be used for the computation.

                                    **Note:** This parameter is honoured if the raster uses "image_server" engine.
    ----------------------------    --------------------------------------------------------------------
    pixel_size                      Optional list or dictionary. The pixel level being used (or the
                                    resolution being looked at). If pixel size is not specified, then
                                    pixel_size will default to the base resolution of the dataset.
                                    The structure of the pixel_size parameter is the same as the
                                    structure of the point object returned by the ArcGIS REST API.
                                    In addition to the dictionary structure, you can specify the pixel size
                                    with a comma-separated string.
                                    
                                    Syntax:
                                    - dictionary structure: pixel_size={point}
                                    - Point simple syntax: pixel_size='<x>,<y>'
                                    Examples:
                                    - pixel_size={"x": 0.18, "y": 0.18}
                                    - pixel_size='0.18,0.18'

                                    **Note:** This parameter is honoured if the raster uses "image_server" engine.
    ----------------------------    --------------------------------------------------------------------
    time                            Optional datetime.date, datetime.datetime or timestamp string. The
                                    time instant or the time extent of the exported image.
                                    Time instant specified as datetime.date, datetime.datetime or
                                    timestamp in milliseconds since epoch
                                    Syntax: time=<timeInstant>
                                    
                                    Time extent specified as list of [<startTime>, <endTime>]
                                    For time extents one of <startTime> or <endTime> could be None. A
                                    None value specified for start time or end time will represent
                                    infinity for start or end time respectively.
                                    Syntax: time=[<startTime>, <endTime>] ; specified as
                                    datetime.date, datetime.datetime or timestamp
                                    
                                    Added at 10.8

                                    **Note:** This parameter is honoured if the raster uses "image_server" engine.
    ----------------------------    --------------------------------------------------------------------
    bands                           Optional list of band indices. By default takes the first band (band index - 0).
                                    Image histogram charts are plotted for these specific bands.

                                    Example:
                                        - [0,2,3]
    ----------------------------    --------------------------------------------------------------------
    display_stats                   Optional boolean. Specifies whether to plot the band-wise statistics 
                                    along with the histograms.

                                    Some basic descriptive statistics are calculated and displayed on 
                                    histograms. The mean and median are displayed with one line each, and 
                                    one standard deviation above and below the mean is displayed using two lines.

                                        - False - The statistics will not be displayed along with the histograms.
                                        - True - The statistics will be displayed along with the histograms. \
                                                This is the default.
    ----------------------------    --------------------------------------------------------------------
    plot_properties                 Optional dictionary. This parameter can be used to set the figure 
                                    properties. These are the `matplotlib.pyplot.figure() <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.figure.html#matplotlib-pyplot-figure>`__ 
                                    parameters and values specified in dict format.

                                    Example:
                                        - {"figsize":(15,15)}
    ----------------------------    --------------------------------------------------------------------
    subplot_properties              Optional list or dictionary. This parameter can be used to set band-wise 
                                    histogram (subplot) display properties. These are the `matplotlib.axes.Axes.bar() <https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.bar.html#matplotlib-axes-axes-bar>`__
                                    parameters and values specified in dictionary format.

                                    Example:
                                        - | [
                                        |  {"color":"r"},
                                        |  {"color":"g"},
                                        |  {"color":"b","edgecolor":"w"}
                                        | ]
                                        
                                    **Note:** `matplotlib.axes.Axes.bar() <https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.bar.html#matplotlib-axes-axes-bar>`__
                                    parameters: ''x', 'height' or 'align' cannot be passed into subplot_properties.
    ============================    ====================================================================

    .. tip::
    When working with multidimensional rasters, you can use the `multidimensional_filter() <https://developers.arcgis.com/python/api-reference/arcgis.raster.functions.html#multidimensional-filter>`__
    raster function on the Raster object for slicing the data along defined variables and dimensions.
    `plot_histograms()` can then be used on the output raster returned upon applying the filter.
    
    :return: None

    """

    subplot_properties_temp = _copy.deepcopy(subplot_properties)
    color_flag = False

    from ._layer import _ArcpyRaster

    if isinstance(raster, _ArcpyRaster):
        try:
            arcpy_raster = raster._raster
            stats = arcpy_raster.getStatistics()
            histograms = arcpy_raster.getHistograms()
            if not histograms or not stats:
                import arcpy

                arcpy.CalculateStatistics_management(arcpy_raster)
            stats = arcpy_raster.getStatistics()
            histograms = arcpy_raster.getHistograms()
        except Exception as e:
            _LOGGER.warning(e)

        if not histograms:
            raise RuntimeError("No histograms found for the raster")
        if not stats:
            raise RuntimeError("No statistics found for the raster")
        stats_histograms = {"statistics": stats, "histograms": histograms}
    else:
        if geometry is None:
            geometry = raster.extent
        stats_histograms = raster.compute_stats_and_histograms(
            geometry=geometry, pixel_size=pixel_size, time=time
        )
        if "histograms" not in stats_histograms or not stats_histograms["histograms"]:
            raise RuntimeError("No histograms found for the raster in the given extent")
        if "statistics" not in stats_histograms or not stats_histograms["statistics"]:
            raise RuntimeError("No statistics found for the raster in the given extent")

    band_count = len(stats_histograms["histograms"])
    bands = [0] if isinstance(bands, list) and len(bands) == 0 else bands
    if isinstance(bands, int) or isinstance(bands, str):
        bands = [bands]
    if isinstance(bands, list):
        for band in bands:
            if not isinstance(band, int) or band not in range(band_count):
                raise RuntimeError("Invalid band index : " + str(band))
    else:
        raise RuntimeError("bands should be of type list")

    if plot_properties is None:
        plot_properties = {}
    if not isinstance(plot_properties, dict):
        raise RuntimeError("plot_properties should be of type dict")
    if len(plot_properties) == 0 or (
        len(plot_properties) > 0 and "figsize" not in plot_properties.keys()
    ):
        plot_properties["figsize"] = (10, 10)

    fig = _plt.figure(**plot_properties)

    if subplot_properties_temp is None:
        subplot_properties_temp = {}
    if isinstance(subplot_properties_temp, list):
        if len(subplot_properties_temp) != len(bands):
            raise RuntimeError(
                "subplot_properties length should be same as the number of band indexes passed into parameter: bands"
            )
    if isinstance(subplot_properties_temp, dict):
        if "color" not in subplot_properties_temp:
            color_flag = True
        subplot_properties_temp = [subplot_properties_temp] * len(bands)

    if isinstance(subplot_properties_temp, list):
        for property_dict in subplot_properties_temp:
            if isinstance(property_dict, dict):
                if any(key in property_dict for key in ["x", "height", "align"]):
                    raise RuntimeError(
                        "subplot_properties dictionaries cannot contain these keys : x, height or align"
                    )
                if len(property_dict) == 0 or "edgecolor" not in property_dict:
                    property_dict.update({"edgecolor": "black"})
            else:
                raise RuntimeError("subplot_properties indexes should be of type dict")
    else:
        raise RuntimeError(
            "subplot_properties should be of type dict or list (of dictionaries)"
        )

    property_gen = iter(subplot_properties_temp)
    color_gen = iter(_cm.rainbow(_np.linspace(0, 1, len(bands))))
    for i, band in enumerate(bands):
        ax = fig.add_subplot(len(bands), 1, i + 1)
        next_property = next(property_gen)
        if "color" not in next_property or color_flag:
            c = next(color_gen)
            next_property["color"] = c
        min_val = stats_histograms["histograms"][band]["min"]
        max_val = stats_histograms["histograms"][band]["max"]
        bins = stats_histograms["histograms"][band]["size"]
        step = (max_val - min_val) / bins
        bin_list = _np.arange(min_val, max_val, step)
        if "width" not in next_property:
            next_property["width"] = step
        next_property.update(
            {
                "x": bin_list,
                "height": stats_histograms["histograms"][band]["counts"],
                "align": "edge",
            }
        )
        ax.bar(**next_property)
        ax.set_xticks(
            _np.linspace(
                stats_histograms["statistics"][band]["min"],
                stats_histograms["statistics"][band]["max"],
                10,
                dtype=int,
            )
        )
        ax.set_ylabel("count")
        ax.set_title("Distribution for Band: " + str(band))

        if display_stats:
            if "mean" in stats_histograms["statistics"][band]:
                mean_stat = stats_histograms["statistics"][band]["mean"]
                ax.axvline(
                    mean_stat,
                    linewidth=1.5,
                    label="Mean: " + str(mean_stat),
                    color="blue",
                )
            if "median" in stats_histograms["statistics"][band]:
                median_stat = stats_histograms["statistics"][band]["median"]
                ax.axvline(
                    median_stat,
                    linewidth=1.5,
                    label="Median: " + str(median_stat),
                    color="green",
                )
            if "standardDeviation" in stats_histograms["statistics"][band]:
                sd_stat = stats_histograms["statistics"][band]["standardDeviation"]
                if "mean" in stats_histograms["statistics"][band]:
                    ax.axvline(
                        mean_stat + sd_stat,
                        linewidth=1.5,
                        label="StdDev: " + str(sd_stat),
                        color="gray",
                        linestyle="--",
                    )
                    ax.axvline(
                        mean_stat - sd_stat, linewidth=1.5, color="gray", linestyle="--"
                    )
        if display_stats or "label" in next_property:
            ax.legend()
    fig.suptitle("Histograms", fontsize=20)
    if "subplotpars" not in plot_properties:
        _plt.subplots_adjust(hspace=0.5)
    _plt.show()


def spectral_profile(
    raster,
    points: list[Point] = [],
    show_values: bool = False,
    plot_properties: dict[str, Any] = {},
):
    """
    Spectral profile charts allow you to select areas of interest or ground features on the image and review the spectral information of all bands in a chart format.

    The x-axis of the spectral profile displays the band names

    The y-axis of the spectral profile displays the spectral values.


    ====================================     ====================================================================
    **Parameter**                             **Description**
    ------------------------------------     --------------------------------------------------------------------
    raster                                   Required Imagery Layer object.
    ------------------------------------     --------------------------------------------------------------------
    points                                   Required list of point Geometry objects.
    ------------------------------------     --------------------------------------------------------------------
    show_values                              Optional bool. Default False.
                                             Set this parameter to True to display the values at each point in the line graph.
    ------------------------------------     --------------------------------------------------------------------
    plot_properties                          Optional dict. This parameter can be used to set the figure
                                             properties. These are the matplotlib.pyplot.figure() parameters and values
                                             specified in dict format.

                                             eg: {"figsize":(15,15)}
    ====================================     ====================================================================

    :return:
        None

    """

    if not isinstance(points, list):
        points = [points]

    num_lines = len(points)
    y = [[] for i in range(0, num_lines)]
    x = [[] for i in range(0, num_lines)]
    # x_var = raster.properties.timeInfo['startTimeField']

    res = []
    t1 = []
    d1 = []
    xx = []
    yy = []
    band_names_list = []
    key_props = raster.key_properties()

    if "BandProperties" in key_props:
        band_properties = key_props["BandProperties"]
    for ele in band_properties:
        if "BandName" in ele.keys():
            band_names_list.append(ele["BandName"])

    if len(band_names_list) == 0:
        if "bandNames" in raster.properties:
            band_names_list = raster.properties.bandNames

    if len(band_names_list) == 0:

        def create_band_names(val):
            val = "Band_" + str(val)
            return val

        if "bandCount" in raster.properties:
            band_count = raster.properties.bandCount
            band_names_list = list(
                map(create_band_names, range(1, len(band_count) + 1))
            )

    for index, point in enumerate(points):
        res = raster.get_samples(
            geometry=point,
            return_first_value_only=False,
            out_fields="*",
        )
        values = res[0]["value"]
        vals = [int(s) for s in values.split(" ")]

        x = band_names_list
        y = vals
        if len(x) != len(y):
            x = x[0 : len(y)]

        t1.append(
            {
                "y": y,
                "x": x,
                "point": index,
            }
        )

    if plot_properties is None:
        plot_properties = {}
    if len(plot_properties) == 0 or (
        len(plot_properties) > 0 and "figsize" not in plot_properties.keys()
    ):
        plot_properties.update({"figsize": (15, 15)})
    if plot_properties is not None and isinstance(plot_properties, dict):
        # {"figsize":(20,10),"dpi":100,"facecolor":"yellow","edgecolor":"blue","linewidth":10.0,"frameon":False}
        _plt.figure(**plot_properties)
    _plt.xlabel("Band Name")
    _plt.ylabel("Value")

    title_string = "Spectral Profile"

    _plt.title(title_string)

    color = iter(_cm.rainbow(_np.linspace(0, 1, len(t1))))
    for i in range(0, len(t1)):
        label_string = "Location " + str(t1[i]["point"])
        c = next(color)
        _plt.plot(t1[i]["x"], t1[i]["y"], c=c, label=label_string)
        _plt.scatter(t1[i]["x"], t1[i]["y"], c=[c])
        _plt.legend(loc="upper left")

        if show_values:
            for x, y in zip(t1[i]["x"], t1[i]["y"]):
                label = "{:.2f}".format(y)
                _plt.annotate(
                    label,  # this is the text
                    (x, y),  # this is the point to label
                    textcoords="offset points",  # how to position the text
                    xytext=(10, 5),  # distance from text to points (x,y)
                    ha="center",
                )
    _plt.xticks(rotation=25)
    _plt.show()
    # plt.legend()


def dimension_profile(
    raster,
    points: list[Point],
    dimension: [str],
    time: _datetime.datetime,
    variables: list[str] = [],
    show_values: bool = False,
    show_trend_line: bool = False,
    plot_properties: dict[str, Any] = {},
):
    """
    Dimension profile chart visualizes change along a vertical dimension, such as depth or height,
    using a multidimensional raster dataset with a z-dimension.
    Dimension Profile is only available for multidimensional datasets that contain a z-dimension.
    Change is plotted in the form of a line graph for a given location and date or time. This allows
    trends in two variables to be displayed and compared simultaneously, while taking into account
    different unit scales.

    The x-axis of the dimension profile displays the values of the variable. Default minimum and
    maximum x-axis bounds are set based on the range of data values represented on the axis.

    The y-axis of the dimension profile displays the vertical dimension value.


    ====================================     ====================================================================
    **Parameter**                             **Description**
    ------------------------------------     --------------------------------------------------------------------
    raster                                   Required Imagery Layer object.
    ------------------------------------     --------------------------------------------------------------------
    points                                   Required list of point Geometry objects.
    ------------------------------------     --------------------------------------------------------------------
    dimension                                Required dimension name. Use this parameter to set the field that
                                             represents the dimension field in the image service.
    ------------------------------------     --------------------------------------------------------------------
    time                                     Required datetime.date, datetime.datetime or timestamp string.
                                             The time slice that will be used for plotting dimension profile.
    ------------------------------------     --------------------------------------------------------------------
    variables                                Required list of variable names.
                                             The Dimension profile chart allows a maximum of two variables to be displayed.
    ------------------------------------     --------------------------------------------------------------------
    show_values                              Optional bool. Default False.
                                             Set this parameter to True to display the values at each point in the line graph.
    ------------------------------------     --------------------------------------------------------------------
    show_trend_line                          Optional string. Default False.
                                             Set this parameter to True to add a linear trend line to the dimension profile chart.
                                             One trend line will be drawn for each location when charting multiple locations,
                                             or each variable when charting multiple variables.
    ------------------------------------     --------------------------------------------------------------------
    plot_properties                          Optional dict. This parameter can be used to set the figure
                                             properties. These are the matplotlib.pyplot.figure() parameters and values
                                             specified in dict format.

                                             eg: {"figsize":(15,15)}
    ====================================     ====================================================================

    :return:
        None

    """

    t1 = []

    if not isinstance(variables, list):
        variables = [variables]

    if len(variables) > 2:
        _LOGGER.warning(
            "Dimension profile chart allows a maximum of two variables to be displayed. Plotting chart for first two variables in list."
        )
        variables = variables[0:2]

    if not isinstance(points, list):
        points = [points]

    if (
        "hasMultidimensions" in raster.properties
        and raster.properties["hasMultidimensions"] == True
    ):
        # add check if variable has StdZ dimension
        mosaic_rule = {
            "mosaicMethod": "esriMosaicAttribute",
            "ascending": True,
            "mosaicOperation": "MT_FIRST",
            "multidimensionalDefinition": [
                {"variableName": "", "dimensionName": dimension}
            ],
        }

        variable_unit = {}
        var_info = raster.multidimensional_info["multidimensionalInfo"]["variables"]
        for var in variables:
            for ele in var_info:
                if ele["name"].lower() == var.lower():
                    if "unit" in ele.keys():
                        variable_unit[ele["name"]] = ele["unit"]

        res = []
        var_plot = {}
        x = []
        y = []
        ax2 = None
        for variable in variables:
            all_plot_points = []
            for index, point in enumerate(points):
                x = []
                y = []
                for md_def in mosaic_rule["multidimensionalDefinition"]:
                    md_def["variableName"] = variable

                    res = raster.get_samples(
                        geometry=point,
                        return_first_value_only=False,
                        out_fields="*",
                        mosaic_rule=mosaic_rule,
                        time=time,
                    )
                    for ele in res:
                        x.append(ele["values"][0])
                        y.append(ele["attributes"][dimension])
                    all_plot_points.append({"y": y, "x": x, "point": index})
            var_plot[variable] = all_plot_points

        x = []
        y = []
        t1 = []
        d1 = []

        if plot_properties is None:
            plot_properties = {}
        if len(plot_properties) == 0 or (
            len(plot_properties) > 0 and "figsize" not in plot_properties.keys()
        ):
            plot_properties.update({"figsize": (20, 10)})
        if plot_properties is not None and isinstance(plot_properties, dict):
            # {"figsize":(20,10),"dpi":100,"facecolor":"yellow","edgecolor":"blue","linewidth":10.0,"frameon":False}
            fig = _plt.figure(**plot_properties)

        ax1 = fig.add_subplot()
        ax1.set_ylabel(dimension)

        if variables[0] in variable_unit.keys():
            ax1.set_xlabel(variables[0] + " (in " + variable_unit[variables[0]] + ")")
        else:
            ax1.set_xlabel(variables[0])

        if len(variables) > 1:
            ax2 = ax1.twiny()
            if variables[1] in variable_unit.keys():
                ax2.set_xlabel(
                    variables[1] + " (in " + variable_unit[variables[1]] + ")"
                )
            else:
                ax2.set_xlabel(variables[1])

        title_string = f"Change in {', '.join(variables)} at {time} over {dimension}"
        _plt.title(title_string)

        # print(t1)
        no_of_colors = (
            2 * len(points) if show_trend_line == False else 4 * (len(points))
        )
        lines = []
        color = iter(_cm.rainbow(_np.linspace(0, 1, no_of_colors)))

        t1 = var_plot[variables[0]]
        for i in range(0, len(t1)):
            label_string = f"Location {str(t1[i]['point'])} - {variables[0]}"

            c = next(color)
            lines += ax1.plot(t1[i]["x"], t1[i]["y"], c=c, label=label_string)
            ax1.scatter(t1[i]["x"], t1[i]["y"], c=[c])

            if show_values:
                for x, y in zip(t1[i]["x"], t1[i]["y"]):
                    label = "{:.2f}".format(x)
                    ax1.annotate(
                        label,  # this is the text
                        (x, y),  # this is the point to label
                        textcoords="offset points",  # how to position the text
                        xytext=(10, 5),  # distance from text to points (x,y)
                        ha="center",
                    )

            if show_trend_line is True:
                c = next(color)
                y1_trend, x_trend = _linear_regression(
                    len(t1[i]["y"]), t1[i]["y"], t1[i]["x"], t1[i]["x"]
                )
                y_trend = t1[i]["y"]
                lines += ax1.plot(
                    x_trend,
                    y_trend,
                    c=c,
                    linestyle="dashed",
                    label=f"Location {str(t1[i]['point'])} - {variables[0]} trend",
                )
                ax1.plot(x_trend, y_trend, c=c, linestyle="dashed")

        if len(variables) > 1:
            t1 = var_plot[variables[1]]
            for i in range(0, len(t1)):
                label_string = f"Location {str(t1[i]['point'])} - {variables[1]}"

                c = next(color)
                lines += ax2.plot(t1[i]["x"], t1[i]["y"], c=c, label=label_string)
                ax2.scatter(t1[i]["x"], t1[i]["y"], c=[c])

                if show_values:
                    for x, y in zip(t1[i]["x"], t1[i]["y"]):
                        label = "{:.2f}".format(x)
                        ax2.annotate(
                            label,
                            (x, y),
                            textcoords="offset points",
                            xytext=(10, 5),
                            ha="center",
                        )

                if show_trend_line is True:
                    c = next(color)
                    y1_trend, x_trend = _linear_regression(
                        len(t1[i]["y"]), t1[i]["y"], t1[i]["x"], t1[i]["x"]
                    )
                    y_trend = t1[i]["y"]
                    lines += ax2.plot(
                        x_trend,
                        y_trend,
                        c=c,
                        linestyle="dashed",
                        label=f"Location {str(t1[i]['point'])} - {variables[1]} trend",
                    )
                    ax2.plot(x_trend, y_trend, c=c, linestyle="dashed")

            ax2.xaxis.grid()

        labels = [line.get_label() for line in lines]
        ax1.legend(lines, labels, loc="center left", bbox_to_anchor=(1, 0.5))
        ax1.xaxis.grid()

        _plt.show()



from dataclasses import dataclass, asdict, field
from typing import Dict, Union, Optional, Any, ClassVar
import pytz


@dataclass(frozen=True)
class RunInterval:
    """
    Set the run interval for the feed.

    ===============     ================================================================================================
    **Parameter**        **Description**
    ---------------     ------------------------------------------------------------------------------------------------
    cron_expression     String. Cron expression that specifies the run interval. You can use the cron generator at the
                        following link to generate a cron expression: `Cron Expression Generator & Explainer
                        <https://www.freeformatter.com/cron-expression-generator-quartz.html>`_.

                        The default is every one minute, represented by the following expression:

                            "``0 * * ? * * *``"
    ---------------     ------------------------------------------------------------------------------------------------
    timezone            String. Run interval timezone to use. The default is: "America/Los_Angeles"

                        .. note::
                            To learn more about time zones, see
                            `List of tz database time zones <https://en.wikipedia.org/wiki/List_of_tz_database_time_zones>`_ page on Wikipedia.
    ===============     ================================================================================================

    :return: `True` if the operation is a success

    .. code-block:: python

        # Usage Example

        feed.run_interval = RunInterval(
            cron_expression="0 * * ? * * *", timezone="America/Los_Angeles",
        )

        # Seconds value must be between 10 and 59
        # Minutes value must be between 1 and 59
        # Hours value must be between 1 and 23

    """

    cron_expression: str
    timezone: str = field(default="America/Los_Angeles")

    def __post_init__(self):
        if self.cron_expression in (None, ""):
            raise ValueError("Cron expression cannot be empty or None")
        elif self.timezone not in pytz.all_timezones:
            raise ValueError("Invalid time zone string")

    def _build(self):
        return {
            "recurrence": {
                "expression": self.cron_expression,
                "timeZone": self.timezone,
            }
        }



"""
The ``arcgis.graph`` module contains classes and functions for working with ArcGIS Knowledge graphs. The available 
functions allow for searching and querying the graph data, and viewing the data model of the graph database. 

Knowledge graphs consist of entities and the relationships between them, each of which can contain properties which describe 
their attributes. The data model of the knowledge graph can show you which entities, relationships, and properties are in 
your database, along with other information about the graph. Performing a search or openCypher query on the graph will 
return results from the database based on the search or query terms provided.

.. note::
    Applications based on ArcGIS API for Python version 2.0.1 can only communicate with knowledge graphs in an ArcGIS Enterprise 
    10.9.1 or 11.0 deployment. ArcGIS Enterprise 11.1 includes breaking changes for knowledge graphs. Only applications based on 
    ArcGIS API for Python version 2.1.0 or later will be able to communicate with knowledge graphs in an Enterprise 11.1 deployment. 
    See the ArcGIS Enterprise Knowledge Server documentation for more details.
"""

from arcgis.graph._service import KnowledgeGraph



from __future__ import annotations
from typing import Any, Iterator
from functools import lru_cache
from arcgis.auth.api import LazyLoader
from arcgis.auth import EsriSession
import logging

json = LazyLoader("json")
requests = LazyLoader("requests")
_arcgis_gis = LazyLoader("arcgis.gis")

_log = logging.getLogger(__name__)


###########################################################################
class RecycleItem:
    """
    This represents a recycled item from the recycling bin.

    .. note::
            This functionality is **only** available to organizations participating
            in the Recycle Bin private beta.

    .. code-block:: python

        # Usage Example:

        >>> gis = GIS(profile="your_online_profile")

        >>> org_user = gis.users.search("gis_user1")[0]
        >>> for r_item in org_user.recyclebin.content:
                print(f"{r_item.properties['title']:15}{r_item.properties['type']:22}{type(r_item)}")

        trees_item1    Service Definition   <class 'arcgis.gis._impl._content_manager._recyclebin.RecycleItem'>
        trees_item1    Feature Service      <class 'arcgis.gis._impl._content_manager._recyclebin.RecycleItem'>
        AR_Counties    Feature Service      <class 'arcgis.gis._impl._content_manager._recyclebin.RecycleItem'>
    """

    _item: _arcgis_gis.Item = None

    # ---------------------------------------------------------------------
    def __init__(self, itemid: str, properties: dict[str, Any], gis: _arcgis_gis.GIS):
        self._itemid: str = itemid
        self._gis: _arcgis_gis.GIS = gis
        self._session: EsriSession = gis.session
        self._properties: dict[str, Any] = properties

    # ---------------------------------------------------------------------
    def __str__(self) -> str:
        return f"< RecycleItem: title:{self.properties['title']} type:{self.properties['type']} >"

    # ---------------------------------------------------------------------
    def __repr__(self) -> str:
        return f"< RecycleItem: title:{self.properties['title']} type:{self.properties['type']} >"

    # ---------------------------------------------------------------------
    @property
    def properties(self) -> dict[str, Any]:
        """
        The properties of the Item in the recycle bin.
        """
        return self._properties

    # ---------------------------------------------------------------------
    def restore(self) -> _arcgis_gis.Item | None:
        """
        Restores the Item from the recycling bin.

        :return: :class:`~arcgis.gis.Item` | None

        .. code-block:: python

            # Usage Example:

            >>> gis = GIS(profile="your_online_profile")

            >>> gis_user = gis.users.me
            >>> deleted_item = list(gis_user.recyclebin.content)[0]
            >>> restored_item = deleted_item.restore()
        """
        url: str = f"{self._gis._public_rest_url}content/users/{self.properties['owner']}/items/{self.properties['id']}/restore"
        params = {
            "f": "json",
        }
        resp: requests.Response = self._session.post(url, data=params)
        resp.raise_for_status()
        data: dict[str, Any] = resp.json()
        if data.get("success"):
            return _arcgis_gis.Item(gis=self._gis, itemid=data.get("itemId"))
        else:
            _log.info(f"Could not restore {data}")

    # ---------------------------------------------------------------------
    def delete(self) -> bool:
        """
        Permanently removes an Item from the organization

        :return: boolean
        """
        url: str = f"{self._gis._public_rest_url}content/users/{self.properties['owner']}/items/{self.properties['id']}/delete"
        params = {"f": "json", "permanentDelete": json.dumps(True)}
        resp: requests.Response = self._session.post(url, data=params)
        resp.raise_for_status()
        data: dict[str, Any] = resp.json()
        if data.get("success"):
            return True
        else:
            _log.error(f"A problem occurred emptying the recycle bin: {data}")
            return False


###########################################################################
class RecycleBin:
    """
    The `RecycleBin` class allows users to manage items they own that were
    deleted.  Users can :meth:`~arcgis.gis._impl._content_manager.RecycleItem.restore`
    or permanently :meth:`~arcgis.gis._impl._content_manager.RecycleItem.delete`
    items from the recycle bin.

    This class is not meant to be initialized directly, but an instance
    is returned by the :attr:`~arcgis.gis.User.recyclebin` property of the
    :class:`~arcgis.gis.User` class. Users can iterate over the
    :attr:`~arcgis.gis._impl._content_manager.RecycleBin.content`.

    .. note::
            This functionality is **only** available to organizations participating
            in the Recycle Bin private beta.

    .. code-block:: python

        # Usage Example:

        >>> gis = GIS(profile="your_online_profile")

        >>> my_recycle_bin = gis.users.me.recyclebin
    """

    _user: _arcgis_gis.User
    _gis: _arcgis_gis.GIS
    _session: EsriSession

    # ---------------------------------------------------------------------
    def __init__(
        self,
        gis: _arcgis_gis.GIS,
        user: _arcgis_gis.User | str | None = None,
    ):
        self._gis: _arcgis_gis.GIS = gis
        self._session: EsriSession = gis.session
        if user is None:
            self._user = self._gis.users.me
        elif isinstance(user, str):
            self._user = self._gis.users.get(user)
        elif isinstance(user, _arcgis_gis.User):
            self._user = user
        else:
            raise ValueError("The `user` parametre must be a str, User or None.")

    # ----------------------------------------------------------------------
    def __str__(self):
        return f"<{self._user.username}'s {self.__class__.__name__}>"

    # ----------------------------------------------------------------------
    def __repr__(self):
        return f"<{self._user.username}'s {self.__class__.__name__}>"

    # ---------------------------------------------------------------------
    @lru_cache(maxsize=5)
    def _supported(self) -> bool:
        """checks if the org supports the recycling bin operations"""
        url: str = f"{self._gis._public_rest_url}portals/self"
        params: dict[str, Any] = {
            "f": "json",
        }
        resp: requests.Response = self._session.get(url, params=params)
        data: dict[str, Any] = resp.json()
        return data.get("recycleBinEnabled", False)

    # ---------------------------------------------------------------------
    @property
    def content(self) -> Iterator[RecycleItem]:
        """
        Lists the content inside the recycling bin.

        :return: Iterator[RecycleItem]

        .. code-block:: python

            # Usage Example:

            >>> gis = GIS(profile="your_online_profile")

            >>> my_user = gis.users.me
            >>> r_bin_content = my_user.recyclebin.content
            >>> type(r_bin_content)

            <class 'generator'>

            >>> for r_item in r_bin_content:
                    print(f"{r_item.properties['title']":15}{r_item.properties['type']}")

            trees_sd        Service Definition
            trees_flc       Feature Service

        """
        if self._supported() == False:
            _log.info("The recyclebin is not supported on this organization.")
            return

        url: str = f"{self._gis._public_rest_url}content/users/{self._user.username}"
        params = {
            "f": "json",
            "foldersContent": json.dumps(True),
            "inRecycleBin": json.dumps(True),
            "start": 1,
            "num": 20,
            "sortField": "modified",
            "sortOrder": "desc",
        }
        resp: requests.Response = self._session.get(url=url, params=params)
        resp.raise_for_status()
        data: dict[str, Any] = resp.json()
        while True:
            for item in data.get("items", []):
                itemid = item.get("id", None)
                if itemid:
                    yield RecycleItem(itemid=itemid, properties=item, gis=self._gis)
            if data.get("nextStart", -1) == -1 or len(data.get("items", [])) <= 0:
                break
            params["start"] = data.get("nextStart", -1)
            resp: requests.Response = self._session.get(url=url, params=params)
            resp.raise_for_status()
            data: dict[str, Any] = resp.json()



import torch
import torch.nn as nn
from ._rand_lanet_utils import Conv1d, Dilated_res_block, batch_preprocess_dict


class SQNRandLANet(nn.Module):
    def __init__(self, config, in_chanls):
        super().__init__()
        self.config = config

        self.fc0 = Conv1d(in_chanls, 8, kernel_size=1, bn=True)

        self.dilated_res_blocks = nn.ModuleList()
        d_in = 8
        for i in range(self.config["num_layers"]):
            d_out = self.config["out_channels"][i]
            self.dilated_res_blocks.append(Dilated_res_block(d_in, d_out))
            d_in = 2 * d_out

        d_out = 2 * sum(self.config["out_channels"])

        self.fc1 = Conv1d(d_out, 256, kernel_size=1, bn=True)
        self.fc2 = Conv1d(256, 128, kernel_size=1, bn=True)
        self.fc3 = Conv1d(128, 64, kernel_size=1, bn=True)
        self.dropout = nn.Dropout(0.5)
        self.logits = Conv1d(
            64,
            self.config["num_classes"],
            kernel_size=1,
            bn=False,
            activation=None,
        )

    def forward(self, end_points):
        # transform input for infrencing
        if not isinstance(end_points, dict):
            device = end_points.device
            end_points = batch_preprocess_dict(
                end_points.cpu(), self.config, is_sqn=True
            )
            for key in end_points:
                if type(end_points[key]) is list:
                    for i in range(len(end_points[key])):
                        end_points[key][i] = end_points[key][i].to(device)
                else:
                    end_points[key] = end_points[key].to(device)

        features = end_points["features"]  # Batch*channel*npoints
        features = self.fc0(features)

        features = features.unsqueeze(dim=3)  # Batch*channel*npoints*1

        # ###########################Encoder############################

        f_encoder_list = []
        for i in range(self.config["num_layers"]):
            f_encoder_i = self.dilated_res_blocks[i](
                features, end_points["xyz"][i], end_points["neigh_idx"][i]
            )

            f_sampled_i = self.random_sample(f_encoder_i, end_points["sub_idx"][i])
            features = f_sampled_i
            f_encoder_list.append(f_sampled_i)

        # ###########################Query Network######################

        query_feature_list = []
        xyz_query = end_points["xyz"][0]  # original points
        for i in range(self.config["num_layers"]):
            xyz_current = end_points["xyz"][i + 1]
            features_current = f_encoder_list[i]
            f_query_feature_i = self.trilinear_interpolation(
                xyz_query, xyz_current, features_current
            )
            query_feature_list.append(f_query_feature_i)

        features_combined = torch.cat(
            query_feature_list, dim=1
        )  # (Batch , 928, npoints)

        features = self.fc1(features_combined)
        features = self.fc2(features)
        features = self.fc3(features)
        features = self.dropout(features)
        f_out = self.logits(features)

        return f_out.permute(0, 2, 1)

    @staticmethod
    def trilinear_interpolation(xyz_query, xyz_support, features_support):
        # Method based on https://github.com/yanx27/Pointnet_Pointnet2_pytorch/blob/master/models/pointnet2_utils.py
        """
        xyz_query : Tensor
            (B, N, 3) tensor of the xyz positions of the unknown points
        xyz_support : Tensor
            (B, M, 3) tensor of the xyz positions of the known points (i.e. B PC examples, each is mx3 shape)
        features_support : Tensor
            (B, C, M, 1) tensor of features to be propagated
        Returns
        new_features : torch.Tensor
            (B, C, N) upsampled tensor
        """
        features_support = features_support.squeeze(dim=3).permute(
            0, 2, 1
        )  # batch*npoints*channel(B, M, C)

        B, N, C = xyz_query.shape
        _, M, _ = xyz_support.shape

        dists = square_distance(xyz_query, xyz_support)
        dists, idx = dists.sort(dim=-1)
        dists, idx = dists[:, :, :3], idx[:, :, :3]  # [B, N, 3]

        dist_recip = 1.0 / (dists + 1e-8)
        norm = torch.sum(dist_recip, dim=2, keepdim=True)
        weight = dist_recip / norm
        interpolated_features = torch.sum(
            index_points(features_support, idx) * weight.view(B, N, 3, 1), dim=2
        )

        interpolated_features = interpolated_features.permute(
            0, 2, 1
        )  # batch*channel*npoints
        return interpolated_features

    @staticmethod
    def random_sample(feature, pool_idx):
        """
        :param feature: [B, N, d] input features matrix
        :param pool_idx: [B, N', max_num] N' < N, N' is the selected position after pooling
        :return: pool_features = [B, N', d] pooled features matrix
        """
        feature = feature.squeeze(dim=3)  # batch*channel*npoints
        num_neigh = pool_idx.shape[-1]
        d = feature.shape[1]
        batch_size = pool_idx.shape[0]
        pool_idx = pool_idx.reshape(batch_size, -1)  # batch*(npoints,nsamples)
        pool_features = torch.gather(
            feature, 2, pool_idx.unsqueeze(1).repeat(1, feature.shape[1], 1)
        )
        pool_features = pool_features.reshape(batch_size, d, -1, num_neigh)
        pool_features = pool_features.max(dim=3, keepdim=True)[
            0
        ]  # batch*channel*npoints*1
        return pool_features


def square_distance(src, dst):
    # Method based on https://github.com/yanx27/Pointnet_Pointnet2_pytorch/blob/master/models/pointnet2_utils.py
    """
    Calculate Euclid distance between each two points.

    src^T * dst = xn * xm + yn * ym + zn * zm；
    sum(src^2, dim=-1) = xn*xn + yn*yn + zn*zn;
    sum(dst^2, dim=-1) = xm*xm + ym*ym + zm*zm;
    dist = (xn - xm)^2 + (yn - ym)^2 + (zn - zm)^2
         = sum(src**2,dim=-1)+sum(dst**2,dim=-1)-2*src^T*dst

    Input:
        src: source points, [B, N, C]
        dst: target points, [B, M, C]
    Output:
        dist: per-point square distance, [B, N, M]
    """
    B, N, _ = src.shape
    _, M, _ = dst.shape
    dist = -2 * torch.matmul(src, dst.permute(0, 2, 1))
    dist += torch.sum(src**2, -1).view(B, N, 1)
    dist += torch.sum(dst**2, -1).view(B, 1, M)
    return dist


def index_points(points, idx):
    # Method based on https://github.com/yanx27/Pointnet_Pointnet2_pytorch/blob/master/models/pointnet2_utils.py
    """
    Input:
        points: input points data, [B, N, C]
        idx: sample index data, [B, S]
    Return:
        new_points:, indexed points data, [B, S, C]
    """
    device = points.device
    B = points.shape[0]
    view_shape = list(idx.shape)
    view_shape[1:] = [1] * (len(view_shape) - 1)
    repeat_shape = list(idx.shape)
    repeat_shape[0] = 1
    batch_indices = (
        torch.arange(B, dtype=torch.long)
        .to(device)
        .view(view_shape)
        .repeat(repeat_shape)
    )
    new_points = points[batch_indices, idx, :]
    return new_points
